# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import sys
import logging
from weakref import ref as weakref_ref
from pyomo.common.pyomo_typing import overload

from pyomo.common.log import is_debug_set
from pyomo.common.deprecation import RenamedClass
from pyomo.common.modeling import NOTSET
from pyomo.common.formatting import tabular_writer
from pyomo.common.timing import ConstructionTimer
from pyomo.common.numeric_types import (
    native_types,
    native_numeric_types,
    check_if_numeric_type,
)

import pyomo.core.expr as EXPR
from pyomo.core.expr.expr_common import _type_check_exception_arg
import pyomo.core.expr.numeric_expr as numeric_expr
from pyomo.core.base.component import ComponentData, ModelComponentFactory
from pyomo.core.base.global_set import UnindexedComponent_index
from pyomo.core.base.indexed_component import (
    IndexedComponent,
    UnindexedComponent_set,
    IndexedComponent_NDArrayMixin,
)
from pyomo.core.expr.numvalue import as_numeric
from pyomo.core.base.initializer import Initializer

logger = logging.getLogger('pyomo.core')


class NamedExpressionData(numeric_expr.NumericValue):
    """An object that defines a generic "named expression".

    This is the base class for both :class:`ExpressionData` and
    :class:`ObjectiveData`.
    """

    # Note: derived classes are expected to declare the _args_ slot
    __slots__ = ()

    EXPRESSION_SYSTEM = EXPR.ExpressionType.NUMERIC
    PRECEDENCE = 0
    ASSOCIATIVITY = EXPR.OperatorAssociativity.NON_ASSOCIATIVE

    def __call__(self, exception=NOTSET):
        """Compute the value of this expression."""
        exception = _type_check_exception_arg(self, exception)
        (arg,) = self.args
        if arg.__class__ in native_types:
            # Note: native_types includes NoneType
            return arg
        return arg(exception=exception)

    def create_node_with_local_data(self, values, classtype=None):
        """
        Construct a simple expression after constructing the
        contained expression.

        This class provides a consistent interface for constructing a
        node, which is used in tree visitor scripts.
        """
        if classtype is None:
            classtype = self.parent_component()._ComponentDataClass
        obj = classtype()
        obj._args_ = values
        return obj

    def is_named_expression_type(self):
        """A boolean indicating whether this in a named expression."""
        return True

    def is_expression_type(self, expression_system=None):
        """A boolean indicating whether this in an expression."""
        return expression_system is None or expression_system == self.EXPRESSION_SYSTEM

    def arg(self, index):
        if index != 0:
            raise KeyError("Invalid index for expression argument: %d" % index)
        return self.args[0]

    @property
    def args(self):
        return self._args_

    def nargs(self):
        return 1

    def _to_string(self, values, verbose, smap):
        if verbose:
            return "%s{%s}" % (str(self), values[0])
        if self.args[0] is None:
            return "%s{None}" % str(self)
        return values[0]

    def clone(self):
        """Return a clone of this expression (no-op)."""
        return self

    def _apply_operation(self, result):
        # This "expression" is a no-op wrapper, so just return the inner
        # result
        return result[0]

    def polynomial_degree(self):
        """A tuple of subexpressions involved in this expressions operation."""
        if self.args[0] is None:
            return None
        return self.expr.polynomial_degree()

    def _compute_polynomial_degree(self, result):
        return result[0]

    def _is_fixed(self, values):
        return values[0]

    # NamedExpressionData should never return False because
    # they can store subexpressions that contain variables
    def is_potentially_variable(self):
        return True

    @property
    def expr(self):
        (arg,) = self.args
        if arg is None:
            return None
        return as_numeric(arg)

    @expr.setter
    def expr(self, value):
        self.set_value(value)

    def set_value(self, expr):
        """Set the expression on this expression."""
        if expr is None or expr.__class__ in native_numeric_types:
            self._args_ = (expr,)
            return
        try:
            if expr.is_numeric_type():
                self._args_ = (expr,)
                return
        except AttributeError:
            if check_if_numeric_type(expr):
                self._args_ = (expr,)
                return
        raise ValueError(
            f"Cannot assign {expr.__class__.__name__} to "
            f"'{self.name}': {self.__class__.__name__} components only "
            "allow numeric expression types."
        )

    def is_constant(self):
        """A boolean indicating whether this expression is constant."""
        # The underlying expression can always be changed
        # so this should never evaluate as constant
        return False

    def is_fixed(self):
        """A boolean indicating whether this expression is fixed."""
        (e,) = self.args
        return e.__class__ in native_types or e.is_fixed()

    # Override the in-place operators here so that we can redirect the
    # dispatcher based on the current contained expression type and not
    # this Expression object (which would map to "other")

    def __iadd__(self, other):
        (e,) = self.args
        return numeric_expr._add_dispatcher[e.__class__, other.__class__](e, other)

    # Note: the default implementation of __isub__ leverages __iadd__
    # and doesn't need to be reimplemented here

    def __imul__(self, other):
        (e,) = self.args
        return numeric_expr._mul_dispatcher[e.__class__, other.__class__](e, other)

    def __idiv__(self, other):
        (e,) = self.args
        return numeric_expr._div_dispatcher[e.__class__, other.__class__](e, other)

    def __itruediv__(self, other):
        (e,) = self.args
        return numeric_expr._div_dispatcher[e.__class__, other.__class__](e, other)

    def __ipow__(self, other):
        (e,) = self.args
        return numeric_expr._pow_dispatcher[e.__class__, other.__class__](e, other)


class _ExpressionData(metaclass=RenamedClass):
    __renamed__new_class__ = NamedExpressionData
    __renamed__version__ = '6.7.2'


class _GeneralExpressionDataImpl(metaclass=RenamedClass):
    __renamed__new_class__ = NamedExpressionData
    __renamed__version__ = '6.7.2'


class ExpressionData(NamedExpressionData, ComponentData):
    """An object that defines an expression that is never cloned

    Parameters
    ----------
    expr : NumericValue
        The Pyomo expression stored in this expression.

    component : Expression
        The Expression object that owns this data.

    """

    __slots__ = ('_args_',)

    def __init__(self, expr=None, component=None):
        self._args_ = (expr,)
        self._component = weakref_ref(component) if (component is not None) else None
        self._index = NOTSET


class _GeneralExpressionData(metaclass=RenamedClass):
    __renamed__new_class__ = ExpressionData
    __renamed__version__ = '6.7.2'


@ModelComponentFactory.register(
    "Named expressions that can be used in other expressions."
)
class Expression(IndexedComponent, IndexedComponent_NDArrayMixin):
    """A shared expression container, which may be defined over an index.

    Parameters
    ----------
    rule : ~.Initializer

        The source to use to initialize the expression(s) in this
        component.  See :func:`.Initializer` for accepted argument types.

    initialize :
        A synonym for `rule`

    expr :
        A synonym for `rule`

    name : str
        Name of this component; will be overridden if this is assigned
        to a Block.

    doc : str
        Text describing this component.

    """

    _ComponentDataClass = ExpressionData
    # This seems like a copy-paste error, and should be renamed/removed
    NoConstraint = IndexedComponent.Skip

    def __new__(cls, *args, **kwds):
        if cls != Expression:
            return super(Expression, cls).__new__(cls)
        if not args or (args[0] is UnindexedComponent_set and len(args) == 1):
            return ScalarExpression.__new__(ScalarExpression)
        else:
            return IndexedExpression.__new__(IndexedExpression)

    @overload
    def __init__(
        self, *indexes, rule=None, expr=None, initialize=None, name=None, doc=None
    ): ...

    def __init__(self, *args, **kwds):
        _init = self._pop_from_kwargs(
            'Expression', kwds, ('rule', 'expr', 'initialize'), None
        )
        # Historically, Expression objects were dense (but None):
        # setting arg_not_specified causes Initializer to recognize
        # _init==None as a constant initializer returning None
        #
        # To initialize a completely empty Expression, pass either
        # initialize={} (to require explicit setitem before a getitem),
        # or initialize=NOTSET (to allow getitem before setitem)
        self._rule = Initializer(_init, arg_not_specified=NOTSET)

        kwds.setdefault('ctype', Expression)
        IndexedComponent.__init__(self, *args, **kwds)

    def _pprint(self):
        return (
            [
                ('Size', len(self)),
                ('Index', None if (not self.is_indexed()) else self._index_set),
            ],
            self.items,
            ("Expression",),
            lambda k, v: ["Undefined" if v.expr is None else v.expr],
        )

    def display(self, prefix="", ostream=None):
        """TODO"""
        if not self.active:
            return
        if ostream is None:
            ostream = sys.stdout
        tab = "    "
        ostream.write(prefix + self.local_name + " : ")
        ostream.write("Size=" + str(len(self)))

        ostream.write("\n")
        tabular_writer(
            ostream,
            prefix + tab,
            ((k, v) for k, v in self._data.items()),
            ("Value",),
            lambda k, v: ["Undefined" if v.expr is None else v()],
        )

    #
    # A utility to extract all index-value pairs defining this
    # expression, returned as a dictionary. useful in many contexts,
    # in which key iteration and repeated __getitem__ calls are too
    # expensive to extract the contents of an expression.
    #
    def extract_values(self):
        return {key: expression_data.expr for key, expression_data in self.items()}

    #
    # takes as input a (index, value) dictionary for updating this
    # Expression.  if check=True, then both the index and value are
    # checked through the __getitem__ method of this class.
    #
    def store_values(self, new_values):
        if (self.is_indexed() is False) and (not None in new_values):
            raise KeyError(
                "Cannot store value for scalar Expression"
                "=" + self.name + "; no value with index "
                "None in input new values map."
            )

        for index, new_value in new_values.items():
            self._data[index].set_value(new_value)

    def _getitem_when_not_present(self, idx):
        if self._rule is None:
            _init = None
            # TBD: Is this desired behavior?  I can see implicitly setting
            # an Expression if it was not originally defined, but I am less
            # convinced that implicitly creating an Expression (like what
            # works with a Var) makes sense.  [JDS 25 Nov 17]
            # raise KeyError(idx)
        else:
            _init = self._rule(self.parent_block(), idx)
            if _init is Expression.Skip:
                raise KeyError(idx)
        return self._setitem_when_not_present(idx, _init)

    def construct(self, data=None):
        """Apply the rule to construct values in this set"""
        if self._constructed:
            return
        self._constructed = True

        timer = ConstructionTimer(self)
        if is_debug_set(logger):
            logger.debug(
                "Constructing Expression, name=%s, from data=%s"
                % (self.name, str(data))
            )

        if self._anonymous_sets is not None:
            for _set in self._anonymous_sets:
                _set.construct()

        try:
            # We do not (currently) accept data for constructing Constraints
            assert data is None
            self._construct_from_rule_using_setitem()
        finally:
            timer.report()


class ScalarExpression(ExpressionData, Expression):
    def __init__(self, *args, **kwds):
        ExpressionData.__init__(self, expr=None, component=self)
        Expression.__init__(self, *args, **kwds)
        self._index = UnindexedComponent_index

    #
    # Override abstract interface methods to first check for
    # construction
    #

    def __call__(self, exception=NOTSET):
        """Return expression on this expression."""
        exception = _type_check_exception_arg(self, exception)
        if self._constructed:
            return super().__call__(exception)
        raise ValueError(
            "Evaluating the expression of Expression '%s' "
            "before the Expression has been constructed (there "
            "is currently no value to return)." % (self.name)
        )

    @property
    def expr(self):
        """Return expression on this expression."""
        if self._constructed:
            return ExpressionData.expr.fget(self)
        raise ValueError(
            "Accessing the expression of Expression '%s' "
            "before the Expression has been constructed (there "
            "is currently no value to return)." % (self.name)
        )

    @expr.setter
    def expr(self, expr):
        """Set the expression on this expression."""
        self.set_value(expr)

    def clear(self):
        self._data = {}

    def set_value(self, expr):
        """Set the expression on this expression."""
        if self._constructed:
            return ExpressionData.set_value(self, expr)
        raise ValueError(
            "Setting the expression of Expression '%s' "
            "before the Expression has been constructed (there "
            "is currently no object to set)." % (self.name)
        )

    def is_constant(self):
        """A boolean indicating whether this expression is constant."""
        if self._constructed:
            return ExpressionData.is_constant(self)
        raise ValueError(
            "Accessing the is_constant flag of Expression '%s' "
            "before the Expression has been constructed (there "
            "is currently no value to return)." % (self.name)
        )

    def is_fixed(self):
        """A boolean indicating whether this expression is fixed."""
        if self._constructed:
            return ExpressionData.is_fixed(self)
        raise ValueError(
            "Accessing the is_fixed flag of Expression '%s' "
            "before the Expression has been constructed (there "
            "is currently no value to return)." % (self.name)
        )

    #
    # Leaving this method for backward compatibility reasons.
    # (probably should be removed)
    #
    def add(self, index, expr):
        """Add an expression with a given index."""
        if index is not None:
            raise KeyError(
                "ScalarExpression object '%s' does not accept "
                "index values other than None. Invalid value: %s" % (self.name, index)
            )
        if (type(expr) is tuple) and (expr == Expression.Skip):
            raise ValueError(
                "Expression.Skip can not be assigned "
                "to an Expression that is not indexed: %s" % (self.name)
            )
        self.set_value(expr)
        return self


class SimpleExpression(metaclass=RenamedClass):
    __renamed__new_class__ = ScalarExpression
    __renamed__version__ = '6.0'


class IndexedExpression(Expression):
    #
    # Leaving this method for backward compatibility reasons
    # Note: It allows adding members outside of self._index_set.
    #       This has always been the case. Not sure there is
    #       any reason to maintain a reference to a separate
    #       index set if we allow this.
    #
    def add(self, index, expr):
        """Add an expression with a given index."""
        if (type(expr) is tuple) and (expr == Expression.Skip):
            return None
        cdata = ExpressionData(expr, component=self)
        self._data[index] = cdata
        return cdata



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import json
from os.path import join, abspath, dirname
import pyomo.contrib.parmest.parmest as parmest
from pyomo.contrib.parmest.examples.semibatch.semibatch import SemiBatchExperiment
import pyomo.contrib.parmest.scenariocreator as sc


def main():

    # Data: list of dictionaries
    data = []
    file_dirname = dirname(abspath(str(__file__)))
    for exp_num in range(10):
        fname = join(file_dirname, 'exp' + str(exp_num + 1) + '.out')
        with open(fname, 'r') as infile:
            d = json.load(infile)
            data.append(d)

    # Create an experiment list
    exp_list = []
    for i in range(len(data)):
        exp_list.append(SemiBatchExperiment(data[i]))

    # View one model
    # exp0_model = exp_list[0].get_labeled_model()
    # exp0_model.pprint()

    pest = parmest.Estimator(exp_list)

    scenmaker = sc.ScenarioCreator(pest, "ipopt")

    # Make one scenario per experiment and write to a csv file
    output_file = "scenarios.csv"
    experimentscens = sc.ScenarioSet("Experiments")
    scenmaker.ScenariosFromExperiments(experimentscens)
    experimentscens.write_csv(output_file)

    # Use the bootstrap to make 3 scenarios and print
    bootscens = sc.ScenarioSet("Bootstrap")
    scenmaker.ScenariosFromBootstrap(bootscens, 3)
    for s in bootscens.ScensIterator():
        print("{}, {}".format(s.name, s.probability))
        for n, v in s.ThetaVals.items():
            print("   {}={}".format(n, v))


if __name__ == "__main__":
    main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from collections import namedtuple
from pyomo.core.base.objective import Objective
from pyomo.common.timing import HierarchicalTimer
from pyomo.common.modeling import unique_component_name
from pyomo.common.config import ConfigBlock, ConfigValue, In
from pyomo.contrib.pynumero.algorithms.solvers.square_solver_base import (
    DenseSquareNlpSolver,
    ScalarDenseSquareNlpSolver,
)
from pyomo.opt import SolverResults, TerminationCondition
from pyomo.common.dependencies import (
    attempt_import,
    numpy as np,
    numpy_available,
    scipy as sp,
    scipy_available,
)

# Use attempt_import here so that we can register the solver even if SciPy is
# not available.
pyomo_nlp, _ = attempt_import("pyomo.contrib.pynumero.interfaces.pyomo_nlp")


class FsolveNlpSolver(DenseSquareNlpSolver):
    OPTIONS = DenseSquareNlpSolver.OPTIONS(
        description="Options for SciPy fsolve wrapper"
    )
    OPTIONS.declare(
        "xtol",
        ConfigValue(
            default=1e-8,
            domain=float,
            description="Tolerance for convergence of variable vector",
        ),
    )
    OPTIONS.declare(
        "maxfev",
        ConfigValue(
            default=100,
            domain=int,
            description="Maximum number of function evaluations per solve",
        ),
    )
    OPTIONS.declare(
        "tol",
        ConfigValue(
            default=None,
            domain=float,
            description="Tolerance for convergence of function residual",
        ),
    )
    OPTIONS.declare("full_output", ConfigValue(default=True, domain=bool))

    def solve(self, x0=None):
        if x0 is None:
            x0 = self._nlp.get_primals()

        res = sp.optimize.fsolve(
            self.evaluate_function,
            x0,
            fprime=self.evaluate_jacobian,
            full_output=self.options.full_output,
            xtol=self.options.xtol,
            maxfev=self.options.maxfev,
        )
        if self.options.full_output:
            x, info, ier, msg = res
        else:
            x, ier, msg = res

        #
        # fsolve converges with a tolerance specified on the variable
        # vector x. We may also want to enforce a tolerance on function
        # value, which we check here.
        #
        if self.options.tol is not None:
            if self.options.full_output:
                fcn_val = info["fvec"]
            else:
                fcn_val = self.evaluate_function(x)
            if not np.all(np.abs(fcn_val) <= self.options.tol):
                raise RuntimeError(
                    "fsolve converged to a solution that does not satisfy the"
                    " function tolerance 'tol' of %s."
                    " You may need to relax the 'tol' option or tighten the"
                    " 'xtol' option (currently 'xtol' is %s)."
                    % (self.options.tol, self.options.xtol)
                )

        return res


class RootNlpSolver(DenseSquareNlpSolver):
    OPTIONS = DenseSquareNlpSolver.OPTIONS(
        description="Options for SciPy fsolve wrapper"
    )
    OPTIONS.declare(
        "tol",
        ConfigValue(default=1e-8, domain=float, description="Convergence tolerance"),
    )
    OPTIONS.declare(
        "method",
        ConfigValue(
            default="hybr",
            domain=In({"hybr", "lm"}),
            description="Method used to solve for the function root",
            doc=("""The 'method' argument in the scipy.optimize.root function.
            For now only 'hybr' (Powell hybrid method from MINPACK) and
            'lm' (Levenberg-Marquardt from MINPACK) are supported.
            """),
        ),
    )

    def solve(self, x0=None):
        if x0 is None:
            x0 = self._nlp.get_primals()

        results = sp.optimize.root(
            self.evaluate_function,
            x0,
            jac=self.evaluate_jacobian,
            tol=self.options.tol,
            method=self.options.method,
        )
        return results


class NewtonNlpSolver(ScalarDenseSquareNlpSolver):
    """A wrapper around the SciPy scalar Newton solver for NLP objects"""

    OPTIONS = ScalarDenseSquareNlpSolver.OPTIONS(
        description="Options for SciPy newton wrapper"
    )
    OPTIONS.declare(
        "tol",
        ConfigValue(default=1e-8, domain=float, description="Convergence tolerance"),
    )
    OPTIONS.declare(
        "secant",
        ConfigValue(
            default=False,
            domain=bool,
            description="Whether to use SciPy's secant method",
        ),
    )
    OPTIONS.declare(
        "full_output",
        ConfigValue(
            default=True,
            domain=bool,
            description="Whether underlying solver should return its full output",
        ),
    )
    OPTIONS.declare(
        "maxiter",
        ConfigValue(
            default=50,
            domain=int,
            description="Maximum number of function evaluations per solve",
        ),
    )

    def solve(self, x0=None):
        if x0 is None:
            x0 = self._nlp.get_primals()

        if self.options.secant:
            fprime = None
        else:
            fprime = lambda x: self.evaluate_jacobian(np.array([x]))[0, 0]
        results = sp.optimize.newton(
            lambda x: self.evaluate_function(np.array([x]))[0],
            x0[0],
            fprime=fprime,
            tol=self.options.tol,
            full_output=self.options.full_output,
            maxiter=self.options.maxiter,
        )
        return results


class SecantNewtonNlpSolver(NewtonNlpSolver):
    """A wrapper around the SciPy scalar Newton solver for NLP objects
    that takes a specified number of secant iterations (default is 2) to
    try to converge a linear equation quickly then switches to Newton's
    method if this is not successful. This strategy is inspired by
    calculate_variable_from_constraint in pyomo.util.calc_var_value.

    """

    OPTIONS = ConfigBlock(description="Options for the SciPy Newton-secant hybrid")
    OPTIONS.declare_from(NewtonNlpSolver.OPTIONS, skip={"maxiter", "secant"})
    OPTIONS.declare(
        "secant_iter",
        ConfigValue(
            default=2,
            domain=int,
            description=(
                "Number of secant iterations to perform before switching"
                " to Newton's method."
            ),
        ),
    )
    OPTIONS.declare(
        "newton_iter",
        ConfigValue(
            default=50,
            domain=int,
            description="Maximum iterations for the Newton solve",
        ),
    )

    def __init__(self, nlp, timer=None, options=None):
        super().__init__(nlp, timer=timer, options=options)
        self.converged_with_secant = None

    def solve(self, x0=None):
        if x0 is None:
            x0 = self._nlp.get_primals()

        try:
            results = sp.optimize.newton(
                lambda x: self.evaluate_function(np.array([x]))[0],
                x0[0],
                fprime=None,
                tol=self.options.tol,
                maxiter=self.options.secant_iter,
                full_output=self.options.full_output,
            )
            self.converged_with_secant = True
        except RuntimeError:
            self.converged_with_secant = False
            x0 = self._nlp.get_primals()
            results = sp.optimize.newton(
                lambda x: self.evaluate_function(np.array([x]))[0],
                x0[0],
                fprime=lambda x: self.evaluate_jacobian(np.array([x]))[0, 0],
                tol=self.options.tol,
                maxiter=self.options.newton_iter,
                full_output=self.options.full_output,
            )
        return results


class PyomoScipySolver:
    def __init__(self, options=None):
        if options is None:
            options = {}
        self._nlp = None
        self._nlp_solver = None
        self._full_output = None
        self.options = options

    def available(self, exception_flag=False):
        return bool(numpy_available and scipy_available)

    def license_is_valid(self):
        return True

    def version(self):
        return tuple(int(_) for _ in sp.__version__.split('.'))

    def set_options(self, options):
        self.options = options

    def solve(self, model, timer=None, tee=False):
        """
        Parameters
        ----------
        model: BlockData
            The model that will be solved
        timer: HierarchicalTimer
            A HierarchicalTimer that "sub-timers" created by this object
            will be attached to. If not provided, a new timer is created.
        tee: Bool
            A dummy flag indicating whether solver output should be displayed.
            The current SciPy solvers supported have no output, so setting this
            flag does not do anything.

        Returns
        -------
        SolverResults
            Contains the results of the solve

        """
        if timer is None:
            timer = HierarchicalTimer()
        self._timer = timer
        self._timer.start("solve")
        active_objs = list(model.component_data_objects(Objective, active=True))
        if len(active_objs) == 0:
            obj_name = unique_component_name(model, "_obj")
            obj = Objective(expr=0.0)
            model.add_component(obj_name, obj)

        nlp = pyomo_nlp.PyomoNLP(model)
        self._nlp = nlp

        if len(active_objs) == 0:
            model.del_component(obj_name)

        # Call to solve(nlp)
        self._nlp_solver = self.create_nlp_solver(options=self.options)
        x0 = nlp.get_primals()
        results = self._nlp_solver.solve(x0=x0)

        # Transfer values back to Pyomo model
        for var, val in zip(nlp.get_pyomo_variables(), nlp.get_primals()):
            var.set_value(val)

        self._timer.stop("solve")

        # Translate results into a Pyomo-compatible results structure
        pyomo_results = self.get_pyomo_results(model, results)

        return pyomo_results

    def get_nlp(self):
        return self._nlp

    def create_nlp_solver(self, **kwds):
        raise NotImplementedError(
            "%s has not implemented the create_nlp_solver method" % self.__class__
        )

    def get_pyomo_results(self, model, scipy_results):
        raise NotImplementedError(
            "%s has not implemented the get_results method" % self.__class__
        )

    #
    # Support "with" statements.
    #
    def __enter__(self):
        return self

    def __exit__(self, t, v, traceback):
        pass


class PyomoFsolveSolver(PyomoScipySolver):
    # Note that scipy.optimize.fsolve does not return a
    # scipy.optimize.OptimizeResult object (as of SciPy 1.9.3).
    # To assess convergence, we must check the integer flag "ier"
    # that is the third (or second if full_output=False) entry
    # of the returned tuple. This dict maps documented "ier" values
    # to Pyomo termination conditions.
    _term_cond = {1: TerminationCondition.feasible}

    def create_nlp_solver(self, **kwds):
        nlp = self.get_nlp()
        solver = FsolveNlpSolver(nlp, **kwds)
        return solver

    def get_pyomo_results(self, model, scipy_results):
        nlp = self.get_nlp()
        if self._nlp_solver.options.full_output:
            x, info, ier, msg = scipy_results
        else:
            x, ier, msg = scipy_results
        results = SolverResults()

        # Record problem data
        results.problem.name = model.name
        results.problem.number_of_constraints = nlp.n_eq_constraints()
        results.problem.number_of_variables = nlp.n_primals()
        results.problem.number_of_binary_variables = 0
        results.problem.number_of_integer_variables = 0
        results.problem.number_of_continuous_variables = nlp.n_primals()

        # Record solver data
        results.solver.name = "scipy.fsolve"
        results.solver.return_code = ier
        results.solver.message = msg
        results.solver.wallclock_time = self._timer.timers["solve"].total_time
        results.solver.termination_condition = self._term_cond.get(
            ier, TerminationCondition.error
        )
        results.solver.status = TerminationCondition.to_solver_status(
            results.solver.termination_condition
        )
        if self._nlp_solver.options.full_output:
            results.solver.number_of_function_evaluations = info["nfev"]
            results.solver.number_of_gradient_evaluations = info["njev"]
        return results


class PyomoRootSolver(PyomoScipySolver):
    def create_nlp_solver(self, **kwds):
        nlp = self.get_nlp()
        solver = RootNlpSolver(nlp, **kwds)
        return solver

    def get_pyomo_results(self, model, scipy_results):
        nlp = self.get_nlp()
        results = SolverResults()

        # Record problem data
        results.problem.name = model.name
        results.problem.number_of_constraints = nlp.n_eq_constraints()
        results.problem.number_of_variables = nlp.n_primals()
        results.problem.number_of_binary_variables = 0
        results.problem.number_of_integer_variables = 0
        results.problem.number_of_continuous_variables = nlp.n_primals()

        # Record solver data
        results.solver.name = "scipy.root"
        results.solver.return_code = scipy_results.status
        results.solver.message = scipy_results.message
        results.solver.wallclock_time = self._timer.timers["solve"].total_time

        # Check the "success" field of the scipy results object as status
        # appears to be different between solvers (i.e. "hybrid" vs "lm")
        # and not well documented as of SciPy 1.9.3
        if scipy_results.success:
            results.solver.termination_condition = TerminationCondition.feasible
        else:
            results.solver.termination_condition = TerminationCondition.error

        results.solver.status = TerminationCondition.to_solver_status(
            results.solver.termination_condition
        )
        # This attribute is in the SciPy documentation but appears not to
        # be implemented for "hybr" or "lm" solvers...
        # results.solver.number_of_iterations = scipy_results.nit
        results.solver.number_of_function_evaluations = scipy_results.nfev
        results.solver.number_of_gradient_evaluations = scipy_results.njev

        return results


class PyomoNewtonSolver(PyomoScipySolver):
    _solver_name = "scipy.newton"

    def create_nlp_solver(self, **kwds):
        nlp = self.get_nlp()
        solver = NewtonNlpSolver(nlp, **kwds)
        return solver

    def get_pyomo_results(self, model, scipy_results):
        nlp = self.get_nlp()
        results = SolverResults()

        if self._nlp_solver.options.full_output:
            root, res = scipy_results
        else:
            root = scipy_results

        # Record problem data
        results.problem.name = model.name
        results.problem.number_of_constraints = nlp.n_eq_constraints()
        results.problem.number_of_variables = nlp.n_primals()
        results.problem.number_of_binary_variables = 0
        results.problem.number_of_integer_variables = 0
        results.problem.number_of_continuous_variables = nlp.n_primals()

        # Record solver data
        results.solver.name = self._solver_name

        results.solver.wallclock_time = self._timer.timers["solve"].total_time

        if self._nlp_solver.options.full_output:
            # We only have access to any of this information if the solver was
            # requested to return its full output.

            # For this solver, res.flag is a string.
            # If successful, it is 'converged'
            results.solver.message = res.flag

            if res.converged:
                term_cond = TerminationCondition.feasible
            else:
                term_cond = TerminationCondition.Error
            results.solver.termination_condition = term_cond
            results.solver.status = TerminationCondition.to_solver_status(
                results.solver.termination_condition
            )

            results.solver.number_of_function_evaluations = res.function_calls
        return results


class PyomoSecantNewtonSolver(PyomoNewtonSolver):
    _solver_name = "scipy.secant-newton"

    def converged_with_secant(self):
        return self._nlp_solver.converged_with_secant

    def create_nlp_solver(self, **kwds):
        nlp = self.get_nlp()
        solver = SecantNewtonNlpSolver(nlp, **kwds)
        return solver



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.contrib.cp.transform.logical_to_disjunctive_walker import (
    LogicalToDisjunctiveVisitor,
)
from pyomo.common.modeling import unique_component_name
from pyomo.common.config import ConfigDict, ConfigValue

from pyomo.core import (
    TransformationFactory,
    VarList,
    Binary,
    LogicalConstraint,
    Block,
    ConstraintList,
    Transformation,
    NonNegativeIntegers,
)
from pyomo.core.base.block import BlockData
from pyomo.core.base import SortComponents
from pyomo.core.util import target_list
from pyomo.gdp import Disjunct, Disjunction


@TransformationFactory.register(
    "contrib.logical_to_disjunctive",
    doc="Convert logical propositions with only Boolean arguments to MIP "
    "representation and convert logical expressions with mixed "
    "integer-Boolean arguments (such as atleast, atmost, and exactly) to "
    "disjunctive representation",
)
class LogicalToDisjunctive(Transformation):
    """
    Re-encode logical constraints as linear constraints,
    converting Boolean variables to binary.
    """

    CONFIG = ConfigDict('core.logical_to_disjunctive')
    CONFIG.declare(
        'targets',
        ConfigValue(
            default=None,
            domain=target_list,
            description="target or list of targets that will be relaxed",
            doc="""
            This specifies the list of LogicalConstraints to transform, or the
            list of Blocks or Disjuncts on which to transform all of the
            LogicalConstraints. Note that if the transformation is done out
            of place, the list of targets should be attached to the model before it
            is cloned, and the list will specify the targets on the cloned
            instance.
            """,
        ),
    )

    def _apply_to(self, model, **kwds):
        config = self.CONFIG(kwds.pop('options', {}))
        config.set_value(kwds)
        targets = config.targets
        if targets is None:
            targets = (model,)

        transBlocks = {}
        visitor = LogicalToDisjunctiveVisitor()
        for t in targets:
            if t.ctype is Block or isinstance(t, BlockData):
                self._transform_block(t, model, visitor, transBlocks)
            elif t.ctype is LogicalConstraint:
                if t.is_indexed():
                    self._transform_constraint(t, visitor, transBlocks)
                else:
                    self._transform_constraintData(t, visitor, transBlocks)
            else:
                raise RuntimeError(
                    "Target '%s' was not a Block, Disjunct, or"
                    " LogicalConstraint. It was of type %s "
                    "and can't be transformed." % (t.name, type(t))
                )

    def _transform_constraint(self, constraint, visitor, transBlocks):
        for i in constraint.keys(sort=SortComponents.ORDERED_INDICES):
            self._transform_constraintData(constraint[i], visitor, transBlocks)
        constraint.deactivate()

    def _transform_block(self, target_block, model, new_varlists, transBlocks):
        _blocks = (
            target_block.values() if target_block.is_indexed() else (target_block,)
        )
        for block in _blocks:
            # Note that this changes the current (though not the original)
            # behavior of logical-to-linear because we descend into Disjuncts in
            # order to find logical constraints. In the context of creating a
            # traditional disjunctive program, this makes sense--we cannot have
            # logical constraints *anywhere* in the active tree after this
            # transformation.
            for logical_constraint in block.component_objects(
                ctype=LogicalConstraint, active=True, descend_into=(Block, Disjunct)
            ):
                self._transform_constraint(
                    logical_constraint, new_varlists, transBlocks
                )

    def _transform_constraintData(self, logical_constraint, visitor, transBlocks):
        # now create a transformation block on the constraint's parent block (if
        # we don't have one already)
        parent_block = logical_constraint.parent_block()
        xfrm_block = transBlocks.get(parent_block)
        if xfrm_block is None:
            xfrm_block = self._create_transformation_block(parent_block)
            transBlocks[parent_block] = xfrm_block

        # This is may be too cute, but just deceive the walker so it puts stuff
        # in the right place.
        visitor.constraints = xfrm_block.transformed_constraints
        visitor.z_vars = xfrm_block.auxiliary_vars
        visitor.disjuncts = xfrm_block.auxiliary_disjuncts
        visitor.disjunctions = xfrm_block.auxiliary_disjunctions
        visitor.walk_expression(logical_constraint.expr)
        logical_constraint.deactivate()

    def _create_transformation_block(self, context):
        new_xfrm_block_name = unique_component_name(context, '_logical_to_disjunctive')
        new_xfrm_block = Block(doc="Transformation objects for logical_to_disjunctive")
        context.add_component(new_xfrm_block_name, new_xfrm_block)

        new_xfrm_block.transformed_constraints = ConstraintList()
        new_xfrm_block.auxiliary_vars = VarList(domain=Binary)
        new_xfrm_block.auxiliary_disjuncts = Disjunct(NonNegativeIntegers)
        new_xfrm_block.auxiliary_disjunctions = Disjunction(NonNegativeIntegers)

        return new_xfrm_block



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from collections.abc import Mapping, MutableMapping

from pyomo.common.autoslots import AutoSlots

from ._hasher import hasher


def _rehash_keys(encode, val):
    if encode:
        return val
    else:
        # object id() may have changed after unpickling,
        # so we rebuild the dictionary keys
        return {hasher[obj.__class__](obj): (obj, v) for obj, v in val.values()}


class ComponentMap(AutoSlots.Mixin, MutableMapping):
    """
    This class is a replacement for dict that allows Pyomo
    modeling components to be used as entry keys. The
    underlying mapping is based on the Python id() of the
    object, which gets around the problem of hashing
    subclasses of NumericValue. This class is meant for
    creating mappings from Pyomo components to values. The
    use of non-Pyomo components as entry keys should be
    avoided.

    A reference to the object is kept around as long as it
    has a corresponding entry in the container, so there is
    no need to worry about id() clashes.

    We also override __setstate__ so that we can rebuild the
    container based on possibly updated object ids after
    a deepcopy or pickle.

    *** An instance of this class should never be
    deepcopied/pickled unless it is done so along with the
    components for which it contains map entries (e.g., as
    part of a block). ***
    """

    __slots__ = ("_dict",)
    __autoslot_mappers__ = {"_dict": _rehash_keys}
    # Expose a "public" interface to the global _hasher dict
    hasher = hasher

    def __init__(self, *args, **kwds):
        # maps id_hash(obj) -> (obj,val)
        self._dict = {}
        # handle the dict-style initialization scenarios
        self.update(*args, **kwds)

    def __str__(self):
        """String representation of the mapping."""
        tmp = {f"{v[0]} (key={k})": v[1] for k, v in self._dict.items()}
        return f"ComponentMap({tmp})"

    #
    # Implement MutableMapping abstract methods
    #

    def __getitem__(self, obj):
        try:
            return self._dict[hasher[obj.__class__](obj)][1]
        except KeyError:
            _id = hasher[obj.__class__](obj)
            raise KeyError(f"{obj} (key={_id})") from None

    def __setitem__(self, obj, val):
        self._dict[hasher[obj.__class__](obj)] = (obj, val)

    def __delitem__(self, obj):
        try:
            del self._dict[hasher[obj.__class__](obj)]
        except KeyError:
            _id = hasher[obj.__class__](obj)
            raise KeyError(f"{obj} (key={_id})") from None

    def __iter__(self):
        return (obj for obj, val in self._dict.values())

    def __len__(self):
        return self._dict.__len__()

    #
    # Overload MutableMapping default implementations
    #

    # We want a specialization of update() to avoid unnecessary calls to
    # id() when copying / merging ComponentMaps
    def update(self, *args, **kwargs):
        if len(args) == 1 and not kwargs and isinstance(args[0], ComponentMap):
            return self._dict.update(args[0]._dict)
        return super().update(*args, **kwargs)

    # We want to avoid generating Pyomo expressions due to comparing the
    # keys, so look up each entry from other in this dict.
    def __eq__(self, other):
        if self is other:
            return True
        if not isinstance(other, Mapping) or len(self) != len(other):
            return False
        # Note we have already verified the dicts are the same size
        for key, val in other.items():
            other_id = hasher[key.__class__](key)
            if other_id not in self._dict:
                return False
            self_val = self._dict[other_id][1]
            # Note: check "is" first to help avoid creation of Pyomo
            # expressions (for the case that the values contain the same
            # Pyomo component)
            if self_val is not val and self_val != val:
                return False
        return True

    def __ne__(self, other):
        return not (self == other)

    #
    # The remaining methods have slow default
    # implementations for MutableMapping. In particular,
    # they rely KeyError catching, which is slow for this
    # class because KeyError messages use fully qualified
    # names.
    #

    def __contains__(self, obj):
        return hasher[obj.__class__](obj) in self._dict

    def clear(self):
        "D.clear() -> None.  Remove all items from D."
        self._dict.clear()

    def get(self, key, default=None):
        "D.get(k[,d]) -> D[k] if k in D, else d.  d defaults to None."
        if key in self:
            return self[key]
        return default

    def setdefault(self, key, default=None):
        "D.setdefault(k[,d]) -> D.get(k,d), also set D[k]=d if k not in D"
        if key in self:
            return self[key]
        else:
            self[key] = default
        return default


class DefaultComponentMap(ComponentMap):
    """A :py:class:`defaultdict` admitting Pyomo Components as keys

    This class is a replacement for defaultdict that allows Pyomo
    modeling components to be used as entry keys. The base
    implementation builds on :py:class:`ComponentMap`.

    """

    __slots__ = ("default_factory",)

    def __init__(self, default_factory=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.default_factory = default_factory

    def __missing__(self, key):
        if self.default_factory is None:
            raise KeyError(key)
        self[key] = ans = self.default_factory()
        return ans

    def __getitem__(self, obj):
        _key = hasher[obj.__class__](obj)
        if _key in self._dict:
            return self._dict[_key][1]
        else:
            return self.__missing__(obj)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# Unit Tests for pyomo.opt.base.OS
#

import json
import os
from os.path import join

import pyomo.common.unittest as unittest

from pyomo.common.fileutils import this_file_dir
from pyomo.common.tempfiles import TempfileManager

from pyomo.opt import (
    TerminationCondition,
    ReaderFactory,
    SolutionStatus,
    SolverStatus,
    check_optimal_termination,
    assert_optimal_termination,
)

currdir = this_file_dir()
deleteFiles = True


class Test(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        import pyomo.environ

    def setUp(self):
        TempfileManager.push()

    def tearDown(self):
        TempfileManager.pop(remove=deleteFiles or self.currentTestPassed())

    def test_factory(self):
        with ReaderFactory("sol") as reader:
            if reader is None:
                raise IOError("Reader 'sol' is not registered")
            soln = reader(join(currdir, "test4_sol.sol"), suffixes=["dual"])
            _test = TempfileManager.create_tempfile('factory.txt')
            soln.write(filename=_test, format='json')
            with (
                open(_test, 'r') as out,
                open(join(currdir, "test4_sol.jsn"), 'r') as txt,
            ):
                self.assertStructuredAlmostEqual(
                    json.load(txt), json.load(out), allow_second_superset=True
                )

    def test_infeasible1(self):
        with ReaderFactory("sol") as reader:
            if reader is None:
                raise IOError("Reader 'sol' is not registered")
            soln = reader(join(currdir, "infeasible1.sol"))
            self.assertEqual(
                soln.solver.termination_condition, TerminationCondition.infeasible
            )
            self.assertEqual(soln.solution.status, SolutionStatus.infeasible)
            self.assertEqual(soln.solver.status, SolverStatus.warning)

            self.assertFalse(check_optimal_termination(soln))

            with self.assertRaises(RuntimeError):
                assert_optimal_termination(soln)

    def test_infeasible2(self):
        with ReaderFactory("sol") as reader:
            if reader is None:
                raise IOError("Reader 'sol' is not registered")
            soln = reader(join(currdir, "infeasible2.sol"))
            self.assertEqual(
                soln.solver.termination_condition, TerminationCondition.infeasible
            )
            self.assertEqual(soln.solution.status, SolutionStatus.infeasible)
            self.assertEqual(soln.solver.status, SolverStatus.warning)

    def test_conopt_optimal(self):
        with ReaderFactory("sol") as reader:
            if reader is None:
                raise IOError("Reader 'sol' is not registered")
            soln = reader(join(currdir, "conopt_optimal.sol"))
            self.assertEqual(
                soln.solver.termination_condition, TerminationCondition.optimal
            )
            self.assertEqual(soln.solution.status, SolutionStatus.optimal)
            self.assertEqual(soln.solver.status, SolverStatus.ok)
            self.assertTrue(check_optimal_termination(soln))
            assert_optimal_termination(soln)

    def test_bad_options(self):
        with ReaderFactory("sol") as reader:
            if reader is None:
                raise IOError("Reader 'sol' is not registered")
            with self.assertRaises(ValueError):
                soln = reader(join(currdir, "bad_options.sol"))

    def test_bad_objno(self):
        with ReaderFactory("sol") as reader:
            if reader is None:
                raise IOError("Reader 'sol' is not registered")
            with self.assertRaises(ValueError):
                soln = reader(join(currdir, "bad_objno.sol"))

    def test_bad_objnoline(self):
        with ReaderFactory("sol") as reader:
            if reader is None:
                raise IOError("Reader 'sol' is not registered")
            with self.assertRaises(ValueError):
                soln = reader(join(currdir, "bad_objnoline.sol"))

    def test_iis_no_variable_values(self):
        with ReaderFactory("sol") as reader:
            if reader is None:
                raise IOError("Reader 'sol' is not registered")
            result = reader(
                join(currdir, "iis_no_variable_values.sol"), suffixes=["iis"]
            )
            soln = result.solution(0)
            self.assertEqual(len(list(soln.variable['v0'].keys())), 1)
            self.assertEqual(soln.variable['v0']['iis'], 1)
            self.assertEqual(len(list(soln.variable['v1'].keys())), 1)
            self.assertEqual(soln.variable['v1']['iis'], 1)
            self.assertEqual(len(list(soln.constraint['c0'].keys())), 1)
            self.assertEqual(soln.constraint['c0']['Iis'], 4)
            import pyomo.kernel as pmo

            m = pmo.block()
            m.v0 = pmo.variable()
            m.v1 = pmo.variable()
            m.c0 = pmo.constraint()
            m.iis = pmo.suffix(direction=pmo.suffix.IMPORT)
            from pyomo.core.expr.symbol_map import SymbolMap

            soln.symbol_map = SymbolMap()
            soln.symbol_map.addSymbol(m.v0, 'v0')
            soln.symbol_map.addSymbol(m.v1, 'v1')
            soln.symbol_map.addSymbol(m.c0, 'c0')
            m.load_solution(soln)
            pmo.pprint(m.iis)
            self.assertEqual(m.iis[m.v0], 1)
            self.assertEqual(m.iis[m.v1], 1)
            self.assertEqual(m.iis[m.c0], 4)


if __name__ == "__main__":
    deleteFiles = False
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

#
# NOTE: deprecated code
#
from pyomo.common.deprecation import deprecated
from pyomo.core import (
    TransformationFactory,
    Constraint,
    Set,
    Var,
    Objective,
    AbstractModel,
    maximize,
)
from pyomo.repn import generate_standard_repn
from pyomo.core.plugins.transform.hierarchy import IsomorphicTransformation
from pyomo.core.plugins.transform.standard_form import StandardForm
from pyomo.core.plugins.transform.util import partial, process_canonical_repn


@TransformationFactory.register("core.lagrangian_dual", doc="Create the LP dual model.")
class DualTransformation(IsomorphicTransformation):
    """Creates a standard form Pyomo model that is equivalent to another
    model

    Options
        dual_constraint_suffix      Defaults to ``_constraint``
        dual_variable_prefix        Defaults to ``p_``
        slack_names                 Defaults to ``auxiliary_slack``
        excess_names                Defaults to ``auxiliary_excess``
        lb_names                    Defaults to ``_lower_bound``
        ub_names                    Defaults to ``_upper_bound``
        pos_suffix                  Defaults to ``_plus``
        neg_suffix                  Defaults to ``_minus``

    """

    @deprecated(
        "Use of the pyomo.duality package is deprecated. There are known bugs "
        "in pyomo.duality, and we do not recommend the use of this code. "
        "Development of dualization capabilities has been shifted to "
        "the Pyomo Adversarial Optimization (PAO) library. Please contact "
        "William Hart for further details (wehart@sandia.gov).",
        version='5.6.2',
    )
    def __init__(self, **kwds):
        kwds['name'] = "linear_dual"
        super(DualTransformation, self).__init__(**kwds)

    def _create_using(self, model, **kwds):
        """
        Transform a model to its Lagrangian dual.
        """

        # Optional naming schemes for dual variables and constraints
        constraint_suffix = kwds.pop("dual_constraint_suffix", "_constraint")
        variable_prefix = kwds.pop("dual_variable_prefix", "p_")

        # Optional naming schemes to pass to StandardForm
        sf_kwds = {}
        sf_kwds["slack_names"] = kwds.pop("slack_names", "auxiliary_slack")
        sf_kwds["excess_names"] = kwds.pop("excess_names", "auxiliary_excess")
        sf_kwds["lb_names"] = kwds.pop("lb_names", "_lower_bound")
        sf_kwds["ub_names"] = kwds.pop("ub_names", "_upper_bound")
        sf_kwds["pos_suffix"] = kwds.pop("pos_suffix", "_plus")
        sf_kwds["neg_suffix"] = kwds.pop("neg_suffix", "_minus")

        # Get the standard form model
        sf_transform = StandardForm()
        sf = sf_transform(model, **sf_kwds)

        # Roughly, parse the objectives and constraints to form A, b, and c of
        #
        # min  c'x
        # s.t. Ax  = b
        #       x >= 0
        #
        # and create a new model from them.

        # We use sparse matrix representations

        # {constraint_name: {variable_name: coefficient}}
        A = _sparse(lambda: _sparse(0))

        # {constraint_name: coefficient}
        b = _sparse(0)

        # {variable_name: coefficient}
        c = _sparse(0)

        # Walk constraints
        for con_name, con_array in sf.component_map(Constraint, active=True).items():
            for con in (con_array[ndx] for ndx in con_array.index_set()):
                # The qualified constraint name
                cname = "%s%s" % (variable_prefix, con.local_name)

                # Process the body of the constraint
                body_terms = process_canonical_repn(generate_standard_repn(con.body))

                # Add a numeric constant to the 'b' vector, if present
                b[cname] -= body_terms.pop(None, 0)

                # Add variable coefficients to the 'A' matrix
                row = _sparse(0)
                for vname, coef in body_terms.items():
                    row["%s%s" % (vname, constraint_suffix)] += coef

                # Process the upper bound of the constraint. We rely on
                # StandardForm to produce equality constraints, thus
                # requiring us only to check the lower bounds.
                lower_terms = process_canonical_repn(generate_standard_repn(con.lower))

                # Add a numeric constant to the 'b' matrix, if present
                b[cname] += lower_terms.pop(None, 0)

                # Add any variables to the 'A' matrix, if present
                for vname, coef in lower_terms.items():
                    row["%s%s" % (vname, constraint_suffix)] -= coef

                A[cname] = row

        # Walk objectives. Multiply all coefficients by the objective's 'sense'
        # to convert maximizing objectives to minimizing ones.
        for obj_name, obj_array in sf.component_map(Objective, active=True).items():
            for obj in (obj_array[ndx] for ndx in obj_array.index_set()):
                # The qualified objective name

                # Process the objective
                terms = process_canonical_repn(generate_standard_repn(obj.expr))

                # Add coefficients
                for name, coef in terms.items():
                    c["%s%s" % (name, constraint_suffix)] += coef * obj_array.sense

        # Form the dual
        dual = AbstractModel()

        # Make constraint index set
        constraint_set_init = []
        for var_name, var_array in sf.component_map(Var, active=True).items():
            for var in (var_array[ndx] for ndx in var_array.index_set()):
                constraint_set_init.append("%s%s" % (var.local_name, constraint_suffix))

        # Make variable index set
        variable_set_init = []
        dual_variable_roots = []
        for con_name, con_array in sf.component_map(Constraint, active=True).items():
            for con in (con_array[ndx] for ndx in con_array.index_set()):
                dual_variable_roots.append(con.local_name)
                variable_set_init.append("%s%s" % (variable_prefix, con.local_name))

        # Create the dual Set and Var objects
        dual.var_set = Set(initialize=variable_set_init)
        dual.con_set = Set(initialize=constraint_set_init)
        dual.vars = Var(dual.var_set)

        # Make the dual constraints
        def constraintRule(A, c, ndx, model):
            return sum(A[v][ndx] * model.vars[v] for v in model.var_set) <= c[ndx]

        dual.cons = Constraint(dual.con_set, rule=partial(constraintRule, A, c))

        # Make the dual objective (maximizing)
        def objectiveRule(b, model):
            return sum(b[v] * model.vars[v] for v in model.var_set)

        dual.obj = Objective(rule=partial(objectiveRule, b), sense=maximize)

        return dual.create()


class _sparse(dict):
    """
    Represents a sparse map. Uses a user-provided value to initialize
    entries. If the default value is a callable object, it is called
    with no arguments.

    Examples

      # Sparse vector
      v = _sparse(0)

      # 2-dimensional sparse matrix
      A = _sparse(lambda: _sparse(0))

    """

    def __init__(self, default, *args, **kwds):
        dict.__init__(self, *args, **kwds)

        if hasattr(default, "__call__"):
            self._default_value = None
            self._default_func = default
        else:
            self._default_value = default
            self._default_func = None

    def __getitem__(self, ndx):
        if ndx in self:
            return dict.__getitem__(self, ndx)
        else:
            if self._default_func is not None:
                return self._default_func()
            else:
                return self._default_value



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.common.unittest as unittest

from pyomo.contrib.pynumero.dependencies import (
    numpy as np,
    numpy_available,
    scipy_available,
)

if not (numpy_available and scipy_available):
    raise unittest.SkipTest(
        "Pynumero needs scipy and numpy to run Sparse intrinsic tests"
    )

from pyomo.contrib.pynumero.sparse import BlockVector
import pyomo.contrib.pynumero as pn


class TestSparseIntrinsics(unittest.TestCase):
    def setUp(self):
        self.v1 = np.array([1.1, 2.2, 3.3])
        self.v2 = np.array([4.4, 5.5, 6.6, 7.7])
        self.v3 = np.array([1.1, 2.2, 3.3]) * 2
        self.v4 = np.array([4.4, 5.5, 6.6, 7.7]) * 2
        self.bv = BlockVector(2)
        self.bv2 = BlockVector(2)
        self.bv.set_blocks([self.v1, self.v2])
        self.bv2.set_blocks([self.v3, self.v4])

    def test_where(self):
        bv = self.bv
        condition = bv >= 4.5
        res = pn.where(condition)[0]
        for bid, blk in enumerate(res):
            self.assertTrue(np.allclose(blk, pn.where(bv.get_block(bid) >= 4.5)))

        flat_condition = condition.flatten()
        res = pn.where(condition, 2.0, 1.0)
        res_flat = pn.where(flat_condition, 2.0, 1.0)
        self.assertTrue(np.allclose(res.flatten(), res_flat))

        res = pn.where(condition, 2.0, np.ones(bv.size))
        res_flat = pn.where(flat_condition, 2.0, np.ones(bv.size))
        self.assertTrue(np.allclose(res.flatten(), res_flat))

        res = pn.where(condition, np.ones(bv.size) * 2.0, 1.0)
        res_flat = pn.where(flat_condition, np.ones(bv.size) * 2.0, 1.0)
        self.assertTrue(np.allclose(res.flatten(), res_flat))

        res = pn.where(condition, np.ones(bv.size) * 2.0, np.ones(bv.size))
        res_flat = pn.where(flat_condition, np.ones(bv.size) * 2.0, np.ones(bv.size))
        self.assertTrue(np.allclose(res.flatten(), res_flat))

        bones = BlockVector(2)
        bones.set_blocks([np.ones(3), np.ones(4)])

        res = pn.where(condition, bones * 2.0, 1.0)
        res_flat = pn.where(flat_condition, np.ones(bv.size) * 2.0, 1.0)
        self.assertTrue(np.allclose(res.flatten(), res_flat))

        res = pn.where(condition, 2.0, bones)
        res_flat = pn.where(flat_condition, 2.0, bones)
        self.assertTrue(np.allclose(res.flatten(), res_flat))

        res = pn.where(condition, np.ones(bv.size) * 2.0, bones)
        res_flat = pn.where(flat_condition, np.ones(bv.size) * 2.0, np.ones(bv.size))
        self.assertTrue(np.allclose(res.flatten(), res_flat))

        res = pn.where(condition, bones * 2.0, np.ones(bv.size))
        res_flat = pn.where(flat_condition, np.ones(bv.size) * 2.0, np.ones(bv.size))
        self.assertTrue(np.allclose(res.flatten(), res_flat))

    def test_isin(self):
        bv = self.bv
        test_bv = BlockVector(2)
        a = np.array([1.1, 3.3])
        b = np.array([5.5, 7.7])
        test_bv.set_block(0, a)
        test_bv.set_block(1, b)

        res = pn.isin(bv, test_bv)
        for bid, blk in enumerate(bv):
            self.assertEqual(blk.size, res.get_block(bid).size)
            res_flat = np.isin(blk, test_bv.get_block(bid))
            self.assertTrue(np.allclose(res.get_block(bid), res_flat))

        c = np.concatenate([a, b])
        res = pn.isin(bv, c)
        for bid, blk in enumerate(bv):
            self.assertEqual(blk.size, res.get_block(bid).size)
            res_flat = np.isin(blk, c)
            self.assertTrue(np.allclose(res.get_block(bid), res_flat))

        res = pn.isin(bv, test_bv, invert=True)
        for bid, blk in enumerate(bv):
            self.assertEqual(blk.size, res.get_block(bid).size)
            res_flat = np.isin(blk, test_bv.get_block(bid), invert=True)
            self.assertTrue(np.allclose(res.get_block(bid), res_flat))

        c = np.concatenate([a, b])
        res = pn.isin(bv, c, invert=True)
        for bid, blk in enumerate(bv):
            self.assertEqual(blk.size, res.get_block(bid).size)
            res_flat = np.isin(blk, c, invert=True)
            self.assertTrue(np.allclose(res.get_block(bid), res_flat))

    # ToDo: try np.copy on a blockvector

    def test_intersect1d(self):
        vv1 = np.array([1.1, 3.3])
        vv2 = np.array([4.4, 7.7])
        bvv = BlockVector(2)
        bvv.set_blocks([vv1, vv2])
        res = pn.intersect1d(self.bv, bvv)
        self.assertIsInstance(res, BlockVector)
        self.assertTrue(np.allclose(res.get_block(0), vv1))
        self.assertTrue(np.allclose(res.get_block(1), vv2))
        vv3 = np.array([1.1, 7.7])
        res = pn.intersect1d(self.bv, vv3)
        self.assertIsInstance(res, BlockVector)
        self.assertTrue(np.allclose(res.get_block(0), np.array([1.1])))
        self.assertTrue(np.allclose(res.get_block(1), np.array([7.7])))
        res = pn.intersect1d(vv3, self.bv)
        self.assertIsInstance(res, BlockVector)
        self.assertTrue(np.allclose(res.get_block(0), np.array([1.1])))
        self.assertTrue(np.allclose(res.get_block(1), np.array([7.7])))

    def test_setdiff1d(self):
        vv1 = np.array([1.1, 3.3])
        vv2 = np.array([4.4, 7.7])
        bvv = BlockVector(2)
        bvv.set_blocks([vv1, vv2])
        res = pn.setdiff1d(self.bv, bvv)
        self.assertIsInstance(res, BlockVector)
        self.assertTrue(np.allclose(res.get_block(0), np.array([2.2])))
        self.assertTrue(np.allclose(res.get_block(1), np.array([5.5, 6.6])))
        vv3 = np.array([1.1, 7.7])
        res = pn.setdiff1d(self.bv, vv3)
        self.assertIsInstance(res, BlockVector)
        self.assertTrue(np.allclose(res.get_block(0), np.array([2.2, 3.3])))
        self.assertTrue(np.allclose(res.get_block(1), np.array([4.4, 5.5, 6.6])))



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.core import TransformationFactory
from pyomo.core.plugins.transform.hierarchy import IsomorphicTransformation
from pyomo.core.plugins.transform.nonnegative_transform import NonNegativeTransformation
from pyomo.core.plugins.transform.equality_transform import EqualityTransform


@TransformationFactory.register(
    "core.standard_form", doc="Create an equivalent LP model in standard form."
)
class StandardForm(IsomorphicTransformation):
    r"""
    Produces a standard-form representation of the model. This form has
    the coefficient matrix (A), the cost vector (c), and the
    constraint vector (b), where the 'standard form' problem is

    .. math::

        \min/\max\ & c'x \\
        s.t.\      &  Ax = b \\
                   & x >= 0

    Options
        slack_names         Default auxiliary_slack
        excess_names        Default auxiliary_excess
        lb_names            Default _lower_bound
        up_names            Default _upper_bound
        pos_suffix          Default _plus
        neg_suffix          Default _neg

    """

    def __init__(self, **kwds):
        kwds['name'] = "standard_form"
        super(StandardForm, self).__init__(**kwds)

    def _create_using(self, model, **kwds):
        """
        Transform a model to standard form
        """

        # Optional naming schemes to pass to EqualityTransform
        eq_kwds = {}
        eq_kwds["slack_names"] = kwds.pop("slack_names", "auxiliary_slack")
        eq_kwds["excess_names"] = kwds.pop("excess_names", "auxiliary_excess")
        eq_kwds["lb_names"] = kwds.pop("lb_names", "_lower_bound")
        eq_kwds["ub_names"] = kwds.pop("ub_names", "_upper_bound")

        # Optional naming schemes to pass to NonNegativeTransformation
        nn_kwds = {}
        nn_kwds["pos_suffix"] = kwds.pop("pos_suffix", "_plus")
        nn_kwds["neg_suffix"] = kwds.pop("neg_suffix", "_minus")

        nonneg = NonNegativeTransformation()
        equality = EqualityTransform()

        # Since NonNegativeTransform introduces new constraints
        # (that aren't equality constraints) we call it first.
        #
        # EqualityTransform introduces new variables, but they are
        # constrained to be nonnegative.
        sf = nonneg(model, **nn_kwds)
        sf = equality(sf, **eq_kwds)

        return sf



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
"""
This file contains an external grey box model representing a simple
reactor design problem described in the Pyomo book.
It is part of the external_grey_box examples with PyNumero.

Note: In this case, this model can be solved using
standard Pyomo constructs (see the Pyomo book), but
this is included as an example of the external grey
box model interface.
"""

import pyomo.environ as pyo
import numpy as np
from scipy.sparse import coo_matrix
from pyomo.contrib.pynumero.interfaces.external_grey_box import ExternalGreyBoxModel


def create_pyomo_reactor_model():
    # this function is here to show what the "ReactorModel" would
    # look like if it was coded directly in Pyomo
    # create the concrete model
    m = pyo.ConcreteModel()

    # set the data (native python data)
    k1 = 5.0 / 6.0  # min^-1
    k2 = 5.0 / 3.0  # min^-1
    k3 = 1.0 / 6000.0  # m^3/(gmol min)
    # caf = 10000.0    # gmol/m^3

    # create the variables
    m.sv = pyo.Var(initialize=1.0, bounds=(0, None))
    m.caf = pyo.Var(initialize=1.0)
    m.ca = pyo.Var(initialize=1.0, bounds=(0, None))
    m.cb = pyo.Var(initialize=1.0, bounds=(0, None))
    m.cc = pyo.Var(initialize=1.0, bounds=(0, None))
    m.cd = pyo.Var(initialize=1.0, bounds=(0, None))
    m.cb_ratio = pyo.Var(initialize=1.0)

    # create the objective
    m.obj = pyo.Objective(expr=m.cb_ratio, sense=pyo.maximize)

    # create the constraints
    m.ca_bal = pyo.Constraint(
        expr=(0 == m.sv * m.caf - m.sv * m.ca - k1 * m.ca - 2.0 * k3 * m.ca**2.0)
    )

    m.cb_bal = pyo.Constraint(expr=(0 == -m.sv * m.cb + k1 * m.ca - k2 * m.cb))

    m.cc_bal = pyo.Constraint(expr=(0 == -m.sv * m.cc + k2 * m.cb))

    m.cd_bal = pyo.Constraint(expr=(0 == -m.sv * m.cd + k3 * m.ca**2.0))

    m.cb_ratio_con = pyo.Constraint(expr=m.cb / (m.ca + m.cc + m.cd) - m.cb_ratio == 0)
    m.cafcon = pyo.Constraint(expr=m.caf == 10000)

    return m


class ReactorModel(ExternalGreyBoxModel):
    def __init__(self, use_exact_derivatives=True):
        self._use_exact_derivatives = use_exact_derivatives

    def input_names(self):
        return ['sv', 'caf', 'ca', 'cb', 'cc', 'cd']

    def equality_constraint_names(self):
        return ['ca_bal', 'cb_bal', 'cc_bal', 'cd_bal']

    def output_names(self):
        return ['cb_ratio']

    def finalize_block_construction(self, pyomo_block):
        # set lower bounds on the variables
        pyomo_block.inputs['sv'].setlb(0)
        pyomo_block.inputs['ca'].setlb(0)
        pyomo_block.inputs['cb'].setlb(0)
        pyomo_block.inputs['cc'].setlb(0)
        pyomo_block.inputs['cd'].setlb(0)

        # initialize the variables
        pyomo_block.inputs['sv'].value = 1
        pyomo_block.inputs['caf'].value = 1
        pyomo_block.inputs['ca'].value = 1
        pyomo_block.inputs['cb'].value = 1
        pyomo_block.inputs['cc'].value = 1
        pyomo_block.inputs['cd'].value = 1
        pyomo_block.outputs['cb_ratio'].value = 1

    def set_input_values(self, input_values):
        self._input_values = list(input_values)

    def evaluate_equality_constraints(self):
        sv = self._input_values[0]
        caf = self._input_values[1]
        ca = self._input_values[2]
        cb = self._input_values[3]
        cc = self._input_values[4]
        cd = self._input_values[5]
        k1 = 5 / 6
        k2 = 5 / 3
        k3 = 1 / 6000
        r = np.zeros(4)
        r[0] = sv * caf + (-sv - k1) * ca - 2 * k3 * ca**2
        r[1] = k1 * ca + (-sv - k2) * cb
        r[2] = k2 * cb - sv * cc
        r[3] = k3 * ca**2 - sv * cd
        return r

    def evaluate_outputs(self):
        ca = self._input_values[2]
        cb = self._input_values[3]
        cc = self._input_values[4]
        cd = self._input_values[5]
        cb_ratio = cb / (ca + cc + cd)
        return np.asarray([cb_ratio], dtype=np.float64)

    def evaluate_jacobian_equality_constraints(self):
        sv = self._input_values[0]
        caf = self._input_values[1]
        ca = self._input_values[2]
        cb = self._input_values[3]
        cc = self._input_values[4]
        cd = self._input_values[5]
        k1 = 5 / 6
        k2 = 5 / 3
        k3 = 1 / 6000

        if self._use_exact_derivatives:
            row = np.zeros(12)
            col = np.zeros(12)
            data = np.zeros(12)
            row[0], col[0], data[0] = (0, 0, caf - ca)
            row[1], col[1], data[1] = (0, 1, sv)
            row[2], col[2], data[2] = (0, 2, -sv - k1 - 4 * k3 * ca)
            row[3], col[3], data[3] = (1, 0, -cb)
            row[4], col[4], data[4] = (1, 2, k1)
            row[5], col[5], data[5] = (1, 3, -sv - k2)
            row[6], col[6], data[6] = (2, 0, -cc)
            row[7], col[7], data[7] = (2, 3, k2)
            row[8], col[8], data[8] = (2, 4, -sv)
            row[9], col[9], data[9] = (3, 0, -cd)
            row[10], col[10], data[10] = (3, 2, 2 * k3 * ca)
            row[11], col[11], data[11] = (3, 5, -sv)
            ret = coo_matrix((data, (row, col)), shape=(4, 6))
            return ret
        else:
            delta = 1e-8
            u0 = np.copy(self._input_values)
            y0 = self.evaluate_equality_constraints()
            jac = np.empty((4, 6))
            u = np.copy(self._input_values)
            for j in range(len(u)):
                # perturb the variables
                u[j] += delta
                self.set_input_values(u)
                yperturb = self.evaluate_equality_constraints()
                jac_col = (yperturb - y0) / delta
                jac[:, j] = jac_col
                u[j] = u0[j]

            # return us back to our starting state
            self.set_input_values(u0)

            # this needs to be a sparse coo_matrix
            # this approach is inefficient, but clear for an example
            row = []
            col = []
            data = []
            for r in range(4):
                for c in range(6):
                    row.append(r)
                    col.append(c)
                    data.append(jac[r, c])
            ret = coo_matrix((data, (row, col)), shape=(4, 6))
            return ret

    def evaluate_jacobian_outputs(self):
        ca = self._input_values[2]
        cb = self._input_values[3]
        cc = self._input_values[4]
        cd = self._input_values[5]
        cb_ratio = cb / (ca + cc + cd)
        row = np.zeros(4)
        col = np.zeros(4)
        data = np.zeros(4)
        row[0], col[0], data[0] = (0, 2, -cb / (ca + cc + cd) ** 2)
        row[1], col[1], data[1] = (0, 3, 1 / (ca + cc + cd))
        row[2], col[2], data[2] = (0, 4, -cb / (ca + cc + cd) ** 2)
        row[3], col[3], data[3] = (0, 5, -cb / (ca + cc + cd) ** 2)
        return coo_matrix((data, (row, col)), shape=(1, 6))


class ReactorModelWithHessian(ReactorModel):
    def __init__(self):
        super(ReactorModelWithHessian, self).__init__(True)
        self._eq_con_mult_values = np.zeros(4)
        self._output_con_mult_values = np.zeros(1)

    def set_equality_constraint_multipliers(self, eq_con_multiplier_values):
        assert len(eq_con_multiplier_values) == 4
        np.copyto(self._eq_con_mult_values, eq_con_multiplier_values)

    def set_output_constraint_multipliers(self, output_con_multiplier_values):
        assert len(output_con_multiplier_values) == 1
        np.copyto(self._output_con_mult_values, output_con_multiplier_values)

    def evaluate_hessian_equality_constraints(self):
        sv = self._input_values[0]
        caf = self._input_values[1]
        ca = self._input_values[2]
        cb = self._input_values[3]
        cc = self._input_values[4]
        cd = self._input_values[5]
        k1 = 5 / 6
        k2 = 5 / 3
        k3 = 1 / 6000
        lam = self._eq_con_mult_values

        nnz = 7
        irow = np.zeros(nnz, dtype=np.int64)
        jcol = np.zeros(nnz, dtype=np.int64)
        data = np.zeros(nnz, dtype=np.float64)
        idx = 0
        irow[idx], jcol[idx], data[idx] = (1, 0, lam[0] * 1.0)
        idx += 1
        irow[idx], jcol[idx], data[idx] = (2, 0, lam[0] * (-1.0))
        idx += 1
        irow[idx], jcol[idx], data[idx] = (2, 2, lam[0] * (-4 * k3))
        idx += 1

        irow[idx], jcol[idx], data[idx] = (3, 0, lam[1] * (-1))
        idx += 1

        irow[idx], jcol[idx], data[idx] = (4, 0, lam[2] * (-1))
        idx += 1

        irow[idx], jcol[idx], data[idx] = (2, 2, lam[3] * (2 * k3))
        idx += 1
        irow[idx], jcol[idx], data[idx] = (5, 0, lam[3] * (-1))
        idx += 1

        assert idx == nnz
        hess = coo_matrix((data, (irow, jcol)), shape=(6, 6))
        return hess

    def evaluate_hessian_outputs(self):
        sv = self._input_values[0]
        caf = self._input_values[1]
        ca = self._input_values[2]
        cb = self._input_values[3]
        cc = self._input_values[4]
        cd = self._input_values[5]
        k1 = 5 / 6
        k2 = 5 / 3
        k3 = 1 / 6000
        lam = self._output_con_mult_values

        nnz = 9
        irow = np.zeros(nnz, dtype=np.int64)
        jcol = np.zeros(nnz, dtype=np.int64)
        data = np.zeros(nnz, dtype=np.float64)

        h1 = 2 * cb / ((ca + cc + cd) ** 3)
        h2 = -1.0 / ((ca + cc + cd) ** 2)

        idx = 0
        irow[idx], jcol[idx], data[idx] = (2, 2, lam[0] * h1)
        idx += 1
        irow[idx], jcol[idx], data[idx] = (3, 2, lam[0] * h2)
        idx += 1
        irow[idx], jcol[idx], data[idx] = (4, 2, lam[0] * h1)
        idx += 1
        irow[idx], jcol[idx], data[idx] = (4, 3, lam[0] * h2)
        idx += 1
        irow[idx], jcol[idx], data[idx] = (4, 4, lam[0] * h1)
        idx += 1
        irow[idx], jcol[idx], data[idx] = (5, 2, lam[0] * h1)
        idx += 1
        irow[idx], jcol[idx], data[idx] = (5, 3, lam[0] * h2)
        idx += 1
        irow[idx], jcol[idx], data[idx] = (5, 4, lam[0] * h1)
        idx += 1
        irow[idx], jcol[idx], data[idx] = (5, 5, lam[0] * h1)
        idx += 1

        assert idx == nnz
        hess = coo_matrix((data, (irow, jcol)), shape=(6, 6))
        return hess


class ReactorModelNoOutputs(ExternalGreyBoxModel):
    def input_names(self):
        return ['sv', 'caf', 'ca', 'cb', 'cc', 'cd']

    def equality_constraint_names(self):
        return ['ca_bal', 'cb_bal', 'cc_bal', 'cd_bal']

    def output_names(self):
        return []

    def finalize_block_construction(self, pyomo_block):
        # set lower bounds on the variables
        pyomo_block.inputs['sv'].setlb(0)
        pyomo_block.inputs['ca'].setlb(0)
        pyomo_block.inputs['cb'].setlb(0)
        pyomo_block.inputs['cc'].setlb(0)
        pyomo_block.inputs['cd'].setlb(0)

        # initialize the variables
        pyomo_block.inputs['sv'].value = 1
        pyomo_block.inputs['caf'].value = 1
        pyomo_block.inputs['ca'].value = 1
        pyomo_block.inputs['cb'].value = 1
        pyomo_block.inputs['cc'].value = 1
        pyomo_block.inputs['cd'].value = 1

    def set_input_values(self, input_values):
        self._input_values = list(input_values)

    def evaluate_equality_constraints(self):
        sv = self._input_values[0]
        caf = self._input_values[1]
        ca = self._input_values[2]
        cb = self._input_values[3]
        cc = self._input_values[4]
        cd = self._input_values[5]
        k1 = 5 / 6
        k2 = 5 / 3
        k3 = 1 / 6000
        r = np.zeros(4)
        r[0] = sv * caf + (-sv - k1) * ca - 2 * k3 * ca**2
        r[1] = k1 * ca + (-sv - k2) * cb
        r[2] = k2 * cb - sv * cc
        r[3] = k3 * ca**2 - sv * cd
        return r

    def evaluate_outputs(self):
        raise NotImplementedError()

    def evaluate_jacobian_equality_constraints(self):
        sv = self._input_values[0]
        caf = self._input_values[1]
        ca = self._input_values[2]
        cb = self._input_values[3]
        cc = self._input_values[4]
        cd = self._input_values[5]
        k1 = 5 / 6
        k2 = 5 / 3
        k3 = 1 / 6000

        row = np.zeros(12)
        col = np.zeros(12)
        data = np.zeros(12)
        row[0], col[0], data[0] = (0, 0, caf - ca)
        row[1], col[1], data[1] = (0, 1, sv)
        row[2], col[2], data[2] = (0, 2, -sv - k1 - 4 * k3 * ca)
        row[3], col[3], data[3] = (1, 0, -cb)
        row[4], col[4], data[4] = (1, 2, k1)
        row[5], col[5], data[5] = (1, 3, -sv - k2)
        row[6], col[6], data[6] = (2, 0, -cc)
        row[7], col[7], data[7] = (2, 3, k2)
        row[8], col[8], data[8] = (2, 4, -sv)
        row[9], col[9], data[9] = (3, 0, -cd)
        row[10], col[10], data[10] = (3, 2, 2 * k3 * ca)
        row[11], col[11], data[11] = (3, 5, -sv)
        ret = coo_matrix((data, (row, col)), shape=(4, 6))
        return ret

    def evaluate_jacobian_outputs(self):
        raise NotImplementedError()


class ReactorModelScaled(ExternalGreyBoxModel):
    def input_names(self):
        return ['sv', 'caf', 'ca', 'cb', 'cc', 'cd']

    def equality_constraint_names(self):
        return ['ca_bal', 'cb_bal', 'cc_bal', 'cd_bal']

    def output_names(self):
        return ['cb_ratio']

    def finalize_block_construction(self, pyomo_block):
        # set lower bounds on the variables
        pyomo_block.inputs['sv'].setlb(0)
        pyomo_block.inputs['ca'].setlb(0)
        pyomo_block.inputs['cb'].setlb(0)
        pyomo_block.inputs['cc'].setlb(0)
        pyomo_block.inputs['cd'].setlb(0)

        # initialize the variables
        pyomo_block.inputs['sv'].value = 1
        pyomo_block.inputs['caf'].value = 1
        pyomo_block.inputs['ca'].value = 1
        pyomo_block.inputs['cb'].value = 1
        pyomo_block.inputs['cc'].value = 1
        pyomo_block.inputs['cd'].value = 1
        pyomo_block.outputs['cb_ratio'].value = 1

        m = pyomo_block.model()
        if not hasattr(m, 'scaling_factor'):
            # add the scaling factor suffix to the model if it is not already declared
            m.scaling_factor = pyo.Suffix(direction=pyo.Suffix.EXPORT)

        m.scaling_factor[pyomo_block.inputs['sv']] = 1.1
        m.scaling_factor[pyomo_block.inputs['caf']] = 1.2
        m.scaling_factor[pyomo_block.inputs['ca']] = 1.3
        m.scaling_factor[pyomo_block.inputs['cb']] = 1.4
        m.scaling_factor[pyomo_block.inputs['cc']] = 1.5
        m.scaling_factor[pyomo_block.inputs['cd']] = 1.6
        m.scaling_factor[pyomo_block.outputs['cb_ratio']] = 1.7

    def get_equality_constraint_scaling_factors(self):
        return np.asarray([0.1, 0.2, 0.3, 0.4])

    def get_output_constraint_scaling_factors(self):
        return np.asarray([10])

    def set_input_values(self, input_values):
        self._input_values = list(input_values)

    def evaluate_equality_constraints(self):
        sv = self._input_values[0]
        caf = self._input_values[1]
        ca = self._input_values[2]
        cb = self._input_values[3]
        cc = self._input_values[4]
        cd = self._input_values[5]
        k1 = 5 / 6
        k2 = 5 / 3
        k3 = 1 / 6000
        r = np.zeros(4)
        r[0] = sv * caf + (-sv - k1) * ca - 2 * k3 * ca**2
        r[1] = k1 * ca + (-sv - k2) * cb
        r[2] = k2 * cb - sv * cc
        r[3] = k3 * ca**2 - sv * cd
        return r

    def evaluate_outputs(self):
        ca = self._input_values[2]
        cb = self._input_values[3]
        cc = self._input_values[4]
        cd = self._input_values[5]
        cb_ratio = cb / (ca + cc + cd)
        return np.asarray([cb_ratio], dtype=np.float64)

    def evaluate_jacobian_equality_constraints(self):
        sv = self._input_values[0]
        caf = self._input_values[1]
        ca = self._input_values[2]
        cb = self._input_values[3]
        cc = self._input_values[4]
        cd = self._input_values[5]
        k1 = 5 / 6
        k2 = 5 / 3
        k3 = 1 / 6000

        row = np.zeros(12)
        col = np.zeros(12)
        data = np.zeros(12)
        row[0], col[0], data[0] = (0, 0, caf - ca)
        row[1], col[1], data[1] = (0, 1, sv)
        row[2], col[2], data[2] = (0, 2, -sv - k1 - 4 * k3 * ca)
        row[3], col[3], data[3] = (1, 0, -cb)
        row[4], col[4], data[4] = (1, 2, k1)
        row[5], col[5], data[5] = (1, 3, -sv - k2)
        row[6], col[6], data[6] = (2, 0, -cc)
        row[7], col[7], data[7] = (2, 3, k2)
        row[8], col[8], data[8] = (2, 4, -sv)
        row[9], col[9], data[9] = (3, 0, -cd)
        row[10], col[10], data[10] = (3, 2, 2 * k3 * ca)
        row[11], col[11], data[11] = (3, 5, -sv)
        ret = coo_matrix((data, (row, col)), shape=(4, 6))
        return ret

    def evaluate_jacobian_outputs(self):
        ca = self._input_values[2]
        cb = self._input_values[3]
        cc = self._input_values[4]
        cd = self._input_values[5]
        cb_ratio = cb / (ca + cc + cd)
        row = np.zeros(4)
        col = np.zeros(4)
        data = np.zeros(4)
        row[0], col[0], data[0] = (0, 2, -cb / (ca + cc + cd) ** 2)
        row[1], col[1], data[1] = (0, 3, 1 / (ca + cc + cd))
        row[2], col[2], data[2] = (0, 4, -cb / (ca + cc + cd) ** 2)
        row[3], col[3], data[3] = (0, 5, -cb / (ca + cc + cd) ** 2)
        return coo_matrix((data, (row, col)), shape=(1, 6))



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common.dependencies import (
    numpy as np,
    numpy_available,
    pandas as pd,
    pandas_available,
    scipy,
    scipy_available,
    matplotlib,
    matplotlib_available,
)

import platform

is_osx = platform.mac_ver()[0] != ''

import pyomo.common.unittest as unittest
import os

import pyomo.contrib.parmest.parmest as parmest
import pyomo.contrib.parmest as parmestbase
import pyomo.environ as pyo

from pyomo.opt import SolverFactory

ipopt_available = SolverFactory('ipopt').available()

from pyomo.common.fileutils import find_library

pynumero_ASL_available = False if find_library('pynumero_ASL') is None else True

testdir = os.path.dirname(os.path.abspath(__file__))


@unittest.skipIf(
    not parmest.parmest_available,
    "Cannot test parmest: required dependencies are missing",
)
@unittest.skipIf(not ipopt_available, "The 'ipopt' command is not available")
class TestSolver(unittest.TestCase):
    def setUp(self):
        pass

    def test_ipopt_solve_with_stats(self):
        from pyomo.contrib.parmest.examples.rooney_biegler.rooney_biegler import (
            rooney_biegler_model,
        )
        from pyomo.contrib.parmest.utils import ipopt_solve_with_stats

        data = pd.DataFrame(
            data=[[1, 8.3], [2, 10.3], [3, 19.0], [4, 16.0], [5, 15.6], [7, 19.8]],
            columns=['hour', 'y'],
        )

        model = rooney_biegler_model(data)
        solver = pyo.SolverFactory('ipopt')
        solver.solve(model)

        status_obj, solved, iters, time, regu = ipopt_solve_with_stats(model, solver)

        self.assertEqual(solved, True)


if __name__ == '__main__':
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# This module was originally developed as part of the PyUtilib project
# Copyright (c) 2008 Sandia Corporation.
# This software is distributed under the BSD License.
# Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
# the U.S. Government retains certain rights in this software.
# ____________________________________________________________________________________

import pickle
import unittest
from pyomo.common.collections import Bunch


class Test(unittest.TestCase):
    def test_Bunch1(self):
        opt = Bunch('a=None c=d e="1 2 3" f=" 5 "', foo=1, bar='x')
        self.assertEqual(opt.ll, None)
        self.assertEqual(opt.a, None)
        self.assertEqual(opt.c, 'd')
        self.assertEqual(opt.e, '1 2 3')
        self.assertEqual(opt.f, 5)
        self.assertEqual(opt.foo, 1)
        self.assertEqual(opt.bar, 'x')
        self.assertEqual(opt['bar'], 'x')
        opt.xx = 1
        opt['yy'] = 2
        self.assertEqual(
            set(opt.keys()), set(['a', 'bar', 'c', 'f', 'foo', 'e', 'xx', 'yy'])
        )
        opt.x = Bunch(a=1, b=2)
        self.assertEqual(
            set(opt.keys()), set(['a', 'bar', 'c', 'f', 'foo', 'e', 'xx', 'yy', 'x'])
        )
        self.assertEqual(
            repr(opt),
            "Bunch(a = None, bar = 'x', c = 'd', e = '1 2 3', f = 5, "
            "foo = 1, x = Bunch(a = 1, b = 2), xx = 1, yy = 2)",
        )
        self.assertEqual(
            str(opt),
            """a: None
bar: 'x'
c: 'd'
e: '1 2 3'
f: 5
foo: 1
x:
    a: 1
    b: 2
xx: 1
yy: 2""",
        )
        opt._name_ = 'BUNCH'
        self.assertEqual(
            set(opt.keys()), set(['a', 'bar', 'c', 'f', 'foo', 'e', 'xx', 'yy', 'x'])
        )
        self.assertEqual(
            repr(opt),
            "Bunch(a = None, bar = 'x', c = 'd', e = '1 2 3', f = 5, "
            "foo = 1, x = Bunch(a = 1, b = 2), xx = 1, yy = 2)",
        )
        self.assertEqual(
            str(opt),
            """a: None
bar: 'x'
c: 'd'
e: '1 2 3'
f: 5
foo: 1
x:
    a: 1
    b: 2
xx: 1
yy: 2""",
        )

        with self.assertRaisesRegex(
            TypeError, r"Bunch\(\) positional arguments must be strings"
        ):
            Bunch(5)

        with self.assertRaisesRegex(
            ValueError,
            r"Bunch\(\) positional arguments must be space "
            "separated strings of form 'key=value', got 'foo'",
        ):
            Bunch('a=5 foo = 6')

    def test_pickle(self):
        o1 = Bunch('a=None c=d e="1 2 3"', foo=1, bar='x')
        s = pickle.dumps(o1)
        o2 = pickle.loads(s)
        self.assertEqual(o1, o2)

    def test_attr_methods(self):
        b = Bunch()
        b.foo = 5
        self.assertEqual(list(b.keys()), ['foo'])
        self.assertEqual(b.foo, 5)
        b._foo = 50
        self.assertEqual(list(b.keys()), ['foo'])
        self.assertEqual(b.foo, 5)
        self.assertEqual(b._foo, 50)

        del b.foo
        self.assertEqual(list(b.keys()), [])
        self.assertEqual(b.foo, None)
        self.assertEqual(b._foo, 50)

        del b._foo
        self.assertEqual(list(b.keys()), [])
        self.assertEqual(b.foo, None)
        with self.assertRaisesRegex(AttributeError, "Unknown attribute '_foo'"):
            b._foo

    def test_item_methods(self):
        b = Bunch()
        b['foo'] = 5
        self.assertEqual(list(b.keys()), ['foo'])
        self.assertEqual(b['foo'], 5)
        b['_foo'] = 50
        self.assertEqual(list(b.keys()), ['foo'])
        self.assertEqual(b['foo'], 5)
        self.assertEqual(b['_foo'], 50)

        del b['foo']
        self.assertEqual(list(b.keys()), [])
        self.assertEqual(b['foo'], None)
        self.assertEqual(b['_foo'], 50)

        del b['_foo']
        self.assertEqual(list(b.keys()), [])
        self.assertEqual(b['foo'], None)
        with self.assertRaisesRegex(AttributeError, "Unknown attribute '_foo'"):
            b['_foo']

        with self.assertRaisesRegex(ValueError, r"Bunch keys must be str \(got int\)"):
            b[5]

        with self.assertRaisesRegex(ValueError, r"Bunch keys must be str \(got int\)"):
            b[5] = 5

        with self.assertRaisesRegex(ValueError, r"Bunch keys must be str \(got int\)"):
            del b[5]

    def test_update(self):
        data = {
            'a': 1,
            'b': [2, {'bb': 3}, [4, {'bbb': 5}]],
            'c': {'cc': 6, 'ccc': {'e': 7}},
            'd': [],
        }

        # Test passing a dict
        b = Bunch()
        b.update(data)
        self.assertEqual(
            repr(b),
            'Bunch(a = 1, '
            'b = [2, Bunch(bb = 3), [4, Bunch(bbb = 5)]], '
            'c = Bunch(cc = 6, ccc = Bunch(e = 7)), '
            'd = [])',
        )

        # Test passing a generator
        b = Bunch()
        b.update(data.items())
        self.assertEqual(
            repr(b),
            'Bunch(a = 1, '
            'b = [2, Bunch(bb = 3), [4, Bunch(bbb = 5)]], '
            'c = Bunch(cc = 6, ccc = Bunch(e = 7)), '
            'd = [])',
        )

        # Test passing a list
        b = Bunch()
        b.update(list(data.items()))
        self.assertEqual(
            repr(b),
            'Bunch(a = 1, '
            'b = [2, Bunch(bb = 3), [4, Bunch(bbb = 5)]], '
            'c = Bunch(cc = 6, ccc = Bunch(e = 7)), '
            'd = [])',
        )

    def test_str(self):
        data = {
            'a': 1,
            'b': [2, {'bb': 3}, [4, {'bbb': 5}]],
            'c': {'cc': 6, 'ccc': {'e': 7}},
            'd': [],
        }

        b = Bunch()
        b.update(data)
        self.assertEqual(
            str(b),
            '''
a: 1
b:
- 2
-
    bb: 3
- [4, Bunch(bbb = 5)]
c:
    cc: 6
    ccc:
        e: 7
d: []
'''.strip(),
        )

    def test_set_name(self):
        b = Bunch()
        self.assertEqual(b._name_, 'Bunch')
        b.set_name('TEST')
        self.assertEqual(b._name_, 'TEST')



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from math import sqrt
from pyomo.common.dependencies import scipy_available
import pyomo.common.unittest as unittest
from pyomo.contrib.piecewise.tests import models
import pyomo.contrib.piecewise.tests.common_tests as ct
from pyomo.core.base import TransformationFactory
from pyomo.core.expr.compare import (
    assertExpressionsEqual,
    assertExpressionsStructurallyEqual,
)
from pyomo.gdp import Disjunct, Disjunction
from pyomo.environ import Constraint, SolverFactory, Var


class TestTransformPiecewiseModelToOuterRepnGDP(unittest.TestCase):
    def check_log_disjunct(self, d, pts, f, substitute_var, x):
        # We can fit both bounds constraints in one constraint, then we have the
        # linear function
        self.assertEqual(len(d.component_map(Constraint)), 2)
        # indicator_var
        self.assertEqual(len(d.component_map(Var)), 1)
        self.assertIsInstance(d.simplex_halfspaces, Constraint)
        self.assertEqual(d.simplex_halfspaces.lower, pts[0])
        self.assertEqual(d.simplex_halfspaces.upper, pts[1])
        self.assertIs(d.simplex_halfspaces.body, x)

        self.assertIsInstance(d.set_substitute, Constraint)
        assertExpressionsEqual(
            self, d.set_substitute.expr, substitute_var == f(x), places=7
        )

    def check_paraboloid_disjunct(self, d, constraint_coefs, f, substitute_var, x1, x2):
        self.assertEqual(len(d.component_map(Constraint)), 2)
        # just indicator_var
        self.assertEqual(len(d.component_map(Var)), 1)
        for i, cons in d.simplex_halfspaces.items():
            coefs = constraint_coefs[i]
            assertExpressionsEqual(
                self, cons.expr, coefs[0] * x1 + coefs[1] * x2 + coefs[2] <= 0, places=6
            )

        self.assertIsInstance(d.set_substitute, Constraint)
        assertExpressionsEqual(
            self, d.set_substitute.expr, substitute_var == f(x1, x2), places=7
        )

    def check_pw_log(self, m):
        ##
        # Check the transformation of the approximation of log(x)
        ##
        z = m.pw_log.get_transformation_var(m.log_expr)
        self.assertIsInstance(z, Var)
        # Now we can use those Vars to check on what the transformation created
        log_block = z.parent_block()
        ct.check_trans_block_structure(self, log_block)

        # Check that all of the Disjuncts have what they should
        self.assertEqual(len(log_block.disjuncts), 3)
        disjuncts_dict = {
            log_block.disjuncts[0]: ((1, 3), m.f1),
            log_block.disjuncts[1]: ((3, 6), m.f2),
            log_block.disjuncts[2]: ((6, 10), m.f3),
        }
        for d, (pts, f) in disjuncts_dict.items():
            self.check_log_disjunct(d, pts, f, log_block.substitute_var, m.x)

        # Check the Disjunction
        self.assertIsInstance(log_block.pick_a_piece, Disjunction)
        self.assertEqual(len(log_block.pick_a_piece.disjuncts), 3)
        for i in range(2):
            self.assertIs(log_block.pick_a_piece.disjuncts[i], log_block.disjuncts[i])

        # And check the substitute Var is in the objective now.
        self.assertIs(m.obj.expr.expr, log_block.substitute_var)

    def check_pw_paraboloid(self, m):
        ##
        # Check the approximation of the transformation of the paraboloid
        ##
        z = m.pw_paraboloid.get_transformation_var(m.paraboloid_expr)
        self.assertIsInstance(z, Var)
        paraboloid_block = z.parent_block()
        ct.check_trans_block_structure(self, paraboloid_block)

        self.assertEqual(len(paraboloid_block.disjuncts), 4)
        disjuncts_dict = {
            # the normal vectors of the faces are normalized when we get
            # them from scipy:
            paraboloid_block.disjuncts[0]: (
                [
                    [sqrt(2) / 2, -sqrt(2) / 2, sqrt(2) / 2],
                    [-1.0, 0.0, 0.0],
                    [0.0, 1.0, -4.0],
                ],
                m.g1,
            ),
            paraboloid_block.disjuncts[1]: (
                [
                    [-sqrt(2) / 2, sqrt(2) / 2, -sqrt(2) / 2],
                    [0.0, -1.0, 1.0],
                    [1.0, 0.0, -3.0],
                ],
                m.g1,
            ),
            paraboloid_block.disjuncts[2]: (
                [
                    [-sqrt(2) / 2, -sqrt(2) / 2, 7 * sqrt(2) / 2],
                    [0.0, 1.0, -7.0],
                    [1.0, 0.0, -3.0],
                ],
                m.g2,
            ),
            paraboloid_block.disjuncts[3]: (
                [
                    [sqrt(2) / 2, sqrt(2) / 2, -7 * sqrt(2) / 2],
                    [-1.0, 0.0, 0.0],
                    [0.0, -1.0, 4.0],
                ],
                m.g2,
            ),
        }
        for d, (constraint_coefs, f) in disjuncts_dict.items():
            self.check_paraboloid_disjunct(
                d, constraint_coefs, f, paraboloid_block.substitute_var, m.x1, m.x2
            )

        # Check the Disjunction
        self.assertIsInstance(paraboloid_block.pick_a_piece, Disjunction)
        self.assertEqual(len(paraboloid_block.pick_a_piece.disjuncts), 4)
        for i in range(3):
            self.assertIs(
                paraboloid_block.pick_a_piece.disjuncts[i],
                paraboloid_block.disjuncts[i],
            )

        # And check the substitute Var is in the objective now.
        self.assertIs(m.indexed_c[0].body.args[0].expr, paraboloid_block.substitute_var)

    @unittest.skipUnless(scipy_available, "Scipy is not available")
    def test_transformation_do_not_descend(self):
        ct.check_transformation_do_not_descend(self, 'contrib.piecewise.outer_repn_gdp')

    def test_transformation_PiecewiseLinearFunction_targets(self):
        ct.check_transformation_PiecewiseLinearFunction_targets(
            self, 'contrib.piecewise.outer_repn_gdp'
        )

    @unittest.skipUnless(scipy_available, "Scipy is not available")
    def test_descend_into_expressions(self):
        ct.check_descend_into_expressions(self, 'contrib.piecewise.outer_repn_gdp')

    @unittest.skipUnless(scipy_available, "Scipy is not available")
    def test_descend_into_expressions_constraint_target(self):
        ct.check_descend_into_expressions_constraint_target(
            self, 'contrib.piecewise.outer_repn_gdp'
        )

    def test_descend_into_expressions_objective_target(self):
        ct.check_descend_into_expressions_objective_target(
            self, 'contrib.piecewise.outer_repn_gdp'
        )

    @unittest.skipUnless(scipy_available, "scipy is not available")
    @unittest.skipUnless(SolverFactory('gurobi').available(), 'Gurobi is not available')
    @unittest.skipUnless(SolverFactory('gurobi').license_is_valid(), 'No license')
    def test_solve_multiple_choice_model(self):
        m = models.make_log_x_model()
        TransformationFactory('contrib.piecewise.multiple_choice').apply_to(m)
        SolverFactory('gurobi').solve(m)

        ct.check_log_x_model_soln(self, m)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

#
# Test the Pyomo command-line interface
#

from itertools import zip_longest
import os
from os.path import abspath, dirname, join

import pyomo.common.unittest as unittest
import pyomo.scripting.convert as convert

from pyomo.common.fileutils import this_file_dir, PYOMO_ROOT_DIR
from pyomo.common.tempfiles import TempfileManager
from pyomo.repn.tests.nl_diff import load_and_compare_nl_baseline
from pyomo.repn.tests.lp_diff import load_and_compare_lp_baseline

currdir = this_file_dir()
scriptdir = join(PYOMO_ROOT_DIR, 'examples', 'pyomo', 'piecewise')

_NL_diff_tol = 1e-9
_LP_diff_tol = 1e-9


class Test(unittest.TestCase):
    def setUp(self):
        self.cwd = os.getcwd()
        TempfileManager.push()
        self.tmpdir = TempfileManager.create_tempdir()
        os.chdir(self.tmpdir)

    def tearDown(self):
        os.chdir(self.cwd)
        TempfileManager.pop()

    def run_convert2nl(self, name):
        return convert.pyomo2nl(
            [
                '--symbolic-solver-labels',
                '--file-determinism',
                '1',
                join(scriptdir, name),
            ]
        )

    def run_convert2lp(self, name):
        return convert.pyomo2lp(['--symbolic-solver-labels', join(scriptdir, name)])

    def test_step_lp(self):
        """Test examples/pyomo/piecewise/step.py"""
        self.run_convert2lp('step.py')
        _test, _base = join(self.tmpdir, 'unknown.lp'), join(currdir, 'step.lp')
        self.assertEqual(*load_and_compare_lp_baseline(_base, _test))

    def test_step_nl(self):
        """Test examples/pyomo/piecewise/step.py"""
        self.run_convert2nl('step.py')
        _test, _base = join(self.tmpdir, 'unknown.nl'), join(currdir, 'step.nl')
        self.assertEqual(*load_and_compare_nl_baseline(_base, _test))

    def test_nonconvex_lp(self):
        """Test examples/pyomo/piecewise/nonconvex.py"""
        self.run_convert2lp('nonconvex.py')
        _test, _base = join(self.tmpdir, 'unknown.lp'), join(currdir, 'nonconvex.lp')
        self.assertEqual(*load_and_compare_lp_baseline(_base, _test))

    def test_nonconvex_nl(self):
        """Test examples/pyomo/piecewise/nonconvex.py"""
        self.run_convert2nl('nonconvex.py')
        _test, _base = join(self.tmpdir, 'unknown.nl'), join(currdir, 'nonconvex.nl')
        self.assertEqual(*load_and_compare_nl_baseline(_base, _test))

    def test_convex_lp(self):
        """Test examples/pyomo/piecewise/convex.py"""
        self.run_convert2lp('convex.py')
        _test, _base = join(self.tmpdir, 'unknown.lp'), join(currdir, 'convex.lp')
        self.assertEqual(*load_and_compare_lp_baseline(_base, _test))

    def test_convex_nl(self):
        """Test examples/pyomo/piecewise/convex.py"""
        self.run_convert2nl('convex.py')
        _test, _base = join(self.tmpdir, 'unknown.nl'), join(currdir, 'convex.nl')
        self.assertEqual(*load_and_compare_nl_baseline(_base, _test))

    def test_indexed_lp(self):
        """Test examples/pyomo/piecewise/indexed.py"""
        self.run_convert2lp('indexed.py')
        _test, _base = join(self.tmpdir, 'unknown.lp'), join(currdir, 'indexed.lp')
        self.assertEqual(*load_and_compare_lp_baseline(_base, _test))

    def test_indexed_nl(self):
        """Test examples/pyomo/piecewise/indexed.py"""
        self.run_convert2nl('indexed.py')
        _base = join(currdir, 'indexed.nl')
        _test = join(self.tmpdir, 'unknown.nl')
        self.assertEqual(*load_and_compare_nl_baseline(_base, _test))

    def test_indexed_nonlinear_nl(self):
        """Test examples/pyomo/piecewise/indexed_nonlinear.py"""
        self.run_convert2nl('indexed_nonlinear.py')
        _test = join(self.tmpdir, 'unknown.nl')
        _base = join(currdir, 'indexed_nonlinear.nl')
        self.assertEqual(*load_and_compare_nl_baseline(_base, _test))


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from collections import Counter

from pyomo.common.dependencies import numpy as np, numpy_available
from pyomo.common import unittest
from pyomo.contrib.alternative_solutions import gurobi_generate_solutions
from pyomo.contrib.appsi.solvers import Gurobi

import pyomo.contrib.alternative_solutions.tests.test_cases as tc
from pyomo.common.log import LoggingIntercept

gurobipy_available = Gurobi().available()


@unittest.skipIf(not gurobipy_available, "Gurobi MIP solver not available")
class TestSolnPoolUnit(unittest.TestCase):
    """
    Cases to cover:

        LP feasibility (for an LP just one solution should be returned since gurobi cannot enumerate over continuous vars)

        Pass at least one solver option to make sure that work, e.g. time limit

        We need a utility to check that a two sets of solutions are the same.
        Maybe this should be an AOS utility since it may be a thing we will want to do often.
    """

    @unittest.skipIf(not numpy_available, "Numpy not installed")
    def test_ip_feasibility(self):
        """
        Enumerate all solutions for an ip: triangle_ip.

        Check that the correct number of alternate solutions are found.
        """
        m = tc.get_triangle_ip()
        results = gurobi_generate_solutions(m, num_solutions=100)
        objectives = [round(result.objective[1], 2) for result in results]
        actual_solns_by_obj = m.num_ranked_solns
        unique_solns_by_obj = [val for val in Counter(objectives).values()]
        np.testing.assert_array_almost_equal(unique_solns_by_obj, actual_solns_by_obj)

    def test_ip_num_solutions_best_effort(self):
        """
        Enumerate solutions for an ip: triangle_ip.
        Test best effort mode in solution pool.

        Check that the correct number of alternate solutions are found.
        """
        m = tc.get_triangle_ip()
        with LoggingIntercept() as LOG:
            results = gurobi_generate_solutions(
                m, num_solutions=8, solver_options={"PoolSearchMode": 1}
            )
        self.assertRegex(
            'Running gurobi_solnpool with PoolSearchMode=1, best effort search may lead to unexpected behavior\n',
            LOG.getvalue(),
        )
        assert len(results) >= 1, 'Need to find some solutions'

    def test_ip_num_solutions_standard_single_solution_solve(self):
        """
        Enumerate solutions for an ip: triangle_ip.
        Test single solve mode in solution pool.

        Check that the correct number of solutions (1) are found.
        This is not the intended use case for this method.
        This is a warning check.
        """
        m = tc.get_triangle_ip()
        with LoggingIntercept() as LOG:
            results = gurobi_generate_solutions(
                m, num_solutions=8, solver_options={"PoolSearchMode": 0}
            )
        self.assertRegex(
            'Running gurobi_solnpool with PoolSearchMode=0, this is single search mode and not the intended use case for gurobi_generate_solutions\n',
            LOG.getvalue(),
        )
        assert len(results) == 1, 'Need to find only 1 solution'

    def test_ip_num_solutions_seeking_one(self):
        """
        Enumerate solutions for an ip: triangle_ip.
        Test case where only one solution is asked for.

        This is not the intended use case for this code.
        This is a warning check.
        """
        m = tc.get_triangle_ip()
        with LoggingIntercept() as LOG:
            results = gurobi_generate_solutions(m, num_solutions=1)
        self.assertRegex(
            'Running alternative_solutions method to find only 1 solution!\n',
            LOG.getvalue(),
        )
        assert len(results) == 1, 'Need to find only 1 solution'

    def test_ip_num_solutions_seeking_zero(self):
        """
        Enumerate solutions for an ip: triangle_ip.
        Test case where zero solutions are asked for to check assert error.
        """
        m = tc.get_triangle_ip()
        with self.assertRaisesRegex(
            AssertionError, "num_solutions must be positive integer"
        ):
            gurobi_generate_solutions(m, num_solutions=0)

    @unittest.skipIf(not numpy_available, "Numpy not installed")
    def test_ip_num_solutions(self):
        """
        Enumerate 8 solutions for an ip: triangle_ip.

        Check that the correct number of alternate solutions are found.
        """
        m = tc.get_triangle_ip()
        results = gurobi_generate_solutions(m, num_solutions=8)
        assert len(results) == 8
        objectives = [round(result.objective[1], 2) for result in results]
        actual_solns_by_obj = [6, 2]
        unique_solns_by_obj = [val for val in Counter(objectives).values()]
        np.testing.assert_array_almost_equal(unique_solns_by_obj, actual_solns_by_obj)

    @unittest.skipIf(not numpy_available, "Numpy not installed")
    def test_mip_feasibility(self):
        """
        Enumerate all solutions for a mip: indexed_pentagonal_pyramid_mip.

        Check that the correct number of alternate solutions are found.
        """
        m = tc.get_indexed_pentagonal_pyramid_mip()
        results = gurobi_generate_solutions(m, num_solutions=100)
        objectives = [round(result.objective[1], 2) for result in results]
        actual_solns_by_obj = m.num_ranked_solns
        unique_solns_by_obj = [val for val in Counter(objectives).values()]
        np.testing.assert_array_almost_equal(unique_solns_by_obj, actual_solns_by_obj)

    @unittest.skipIf(not numpy_available, "Numpy not installed")
    def test_mip_rel_feasibility(self):
        """
        Enumerate solutions for a mip: indexed_pentagonal_pyramid_mip.

        Check that only solutions within a relative tolerance of 0.2 are
        found.
        """
        m = tc.get_pentagonal_pyramid_mip()
        results = gurobi_generate_solutions(m, num_solutions=100, rel_opt_gap=0.2)
        objectives = [round(result.objective[1], 2) for result in results]
        actual_solns_by_obj = m.num_ranked_solns[0:2]
        unique_solns_by_obj = [val for val in Counter(objectives).values()]
        np.testing.assert_array_almost_equal(unique_solns_by_obj, actual_solns_by_obj)

    @unittest.skipIf(not numpy_available, "Numpy not installed")
    def test_mip_rel_feasibility_options(self):
        """
        Enumerate solutions for a mip: indexed_pentagonal_pyramid_mip.

        Check that only solutions within a relative tolerance of 0.2 are
        found.
        """
        m = tc.get_pentagonal_pyramid_mip()
        results = gurobi_generate_solutions(
            m, num_solutions=100, solver_options={"PoolGap": 0.2}
        )
        objectives = [round(result.objective[1], 2) for result in results]
        actual_solns_by_obj = m.num_ranked_solns[0:2]
        unique_solns_by_obj = [val for val in Counter(objectives).values()]
        np.testing.assert_array_almost_equal(unique_solns_by_obj, actual_solns_by_obj)

    @unittest.skipIf(not numpy_available, "Numpy not installed")
    def test_mip_abs_feasibility(self):
        """
        Enumerate solutions for a mip: indexed_pentagonal_pyramid_mip.

        Check that only solutions within an absolute tolerance of 1.99 are
        found.
        """
        m = tc.get_pentagonal_pyramid_mip()
        results = gurobi_generate_solutions(m, num_solutions=100, abs_opt_gap=1.99)
        objectives = [round(result.objective[1], 2) for result in results]
        actual_solns_by_obj = m.num_ranked_solns[0:3]
        unique_solns_by_obj = [val for val in Counter(objectives).values()]
        np.testing.assert_array_almost_equal(unique_solns_by_obj, actual_solns_by_obj)

    @unittest.skipIf(True, "Ignoring fragile test for solver timeout.")
    def test_mip_no_time(self):
        """
        Enumerate solutions for a mip: indexed_pentagonal_pyramid_mip.

        Check that no solutions are returned with a timelimit of 0.
        """
        m = tc.get_pentagonal_pyramid_mip()
        # Use quiet=False to test error message
        results = gurobi_generate_solutions(
            m, num_solutions=100, solver_options={"TimeLimit": 0.0}, quiet=False
        )
        assert len(results) == 0


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# Unit Tests for Piecewise
#

import os
from os.path import abspath, dirname

currdir = dirname(abspath(__file__)) + os.sep

import pyomo.common.unittest as unittest

from pyomo.environ import AbstractModel, ConcreteModel, Set, Var, Piecewise, Constraint


class TestMiscPiecewise(unittest.TestCase):
    # test that activate and deactivate work
    def test_activate_deactivate_indexed(self):
        model = ConcreteModel()
        model.s = Set(initialize=[1])
        model.y = Var(model.s)
        model.x = Var(model.s, bounds=(-1, 1))
        args = ([1], model.y, model.x)
        keywords = {
            'pw_pts': {1: [-1, 0, 1]},
            'pw_constr_type': 'EQ',
            'f_rule': lambda model, i, x: x**2,
        }
        model.c = Piecewise(*args, **keywords)
        self.assertTrue(len(model.c[1].component_map(Constraint)) > 0)
        self.assertTrue(len(model.c[1].component_map(Constraint, active=True)) > 0)
        self.assertEqual(model.c.active, True)
        self.assertEqual(model.c[1].active, True)
        model.c[1].deactivate()
        self.assertTrue(len(model.c[1].component_map(Constraint)) > 0)
        self.assertTrue(len(model.c[1].component_map(Constraint, active=True)) > 0)
        self.assertEqual(model.c.active, True)
        self.assertEqual(model.c[1].active, False)
        model.c[1].activate()
        self.assertTrue(len(model.c[1].component_map(Constraint)) > 0)
        self.assertTrue(len(model.c[1].component_map(Constraint, active=True)) > 0)
        self.assertEqual(model.c.active, True)
        self.assertEqual(model.c[1].active, True)
        model.c.deactivate()
        self.assertTrue(len(model.c[1].component_map(Constraint)) > 0)
        self.assertTrue(len(model.c[1].component_map(Constraint, active=True)) > 0)
        self.assertEqual(model.c.active, False)
        self.assertEqual(model.c[1].active, False)

    # test that activate and deactivate work
    def test_activate_deactivate_nonindexed(self):
        model = ConcreteModel()
        model.y = Var()
        model.x = Var(bounds=(-1, 1))
        args = (model.y, model.x)
        keywords = {
            'pw_pts': [-1, 0, 1],
            'pw_constr_type': 'EQ',
            'f_rule': lambda model, x: x**2,
        }
        model.c = Piecewise(*args, **keywords)
        self.assertTrue(len(model.c.component_map(Constraint)) > 0)
        self.assertTrue(len(model.c.component_map(Constraint, active=True)) > 0)
        self.assertEqual(model.c.active, True)
        model.c.deactivate()
        self.assertTrue(len(model.c.component_map(Constraint)) > 0)
        self.assertTrue(len(model.c.component_map(Constraint, active=True)) > 0)
        self.assertEqual(model.c.active, False)
        model.c.activate()
        self.assertTrue(len(model.c.component_map(Constraint)) > 0)
        self.assertTrue(len(model.c.component_map(Constraint, active=True)) > 0)
        self.assertEqual(model.c.active, True)

    # test that indexed Piecewise can handle
    # non-indexed vars
    def test_indexed_with_nonindexed_vars(self):
        model = ConcreteModel()
        model.range1 = Var()

        model.x = Var(bounds=(-1, 1))
        args = ([1], model.range1, model.x)
        keywords = {
            'pw_pts': {1: [-1, 0, 1]},
            'pw_constr_type': 'EQ',
            'f_rule': lambda model, i, x: x**2,
        }
        model.con1 = Piecewise(*args, **keywords)

        model.range2 = Var([1])
        model.y = Var([1], bounds=(-1, 1))
        args = ([1], model.range2, model.y)
        model.con2 = Piecewise(*args, **keywords)

        args = ([1], model.range2, model.y[1])
        model.con3 = Piecewise(*args, **keywords)

    # test that nonindexed Piecewise can handle
    # VarData (e.g model.x[1]
    def test_nonindexed_with_indexed_vars(self):
        model = ConcreteModel()
        model.range = Var([1])

        model.x = Var([1], bounds=(-1, 1))
        args = (model.range[1], model.x[1])
        keywords = {
            'pw_pts': [-1, 0, 1],
            'pw_constr_type': 'EQ',
            'f_rule': lambda model, x: x**2,
        }
        model.con1 = Piecewise(*args, **keywords)

    # test that Piecewise can be initialized on
    # an abstract model
    def test_abstract_piecewise(self):
        model = AbstractModel()
        model.range = Var()

        model.x = Var(bounds=(-1, 1))
        args = (model.range, model.x)
        keywords = {
            'pw_pts': [-1, 0, 1],
            'pw_constr_type': 'EQ',
            'f_rule': lambda model, x: x**2,
        }
        model.con = Piecewise(*args, **keywords)
        instance = model.create_instance()

    # test that Piecewise can be initialized on
    # a concrete model
    def test_concrete_piecewise(self):
        model = ConcreteModel()
        model.range = Var()

        model.x = Var(bounds=(-1, 1))
        args = (model.range, model.x)
        keywords = {
            'pw_pts': [-1, 0, 1],
            'pw_constr_type': 'EQ',
            'f_rule': lambda model, x: x**2,
        }
        model.con = Piecewise(*args, **keywords)


class TestInvalidPiecewise(unittest.TestCase):
    # test the that Piecewise component raises
    # an exception when the LOG/DLOG reps
    # are requested without a (2^n)+1 length
    # pw_pts list
    def test_dlog_bad_length(self):
        model = ConcreteModel()
        model.range = Var()

        model.x = Var(bounds=(-1, 1))
        args = (model.range, model.x)
        keywords = {
            'pw_pts': [-1, 0, 1],
            'pw_constr_type': 'EQ',
            'pw_repn': 'DLOG',
            'f_rule': lambda model, x: x**2,
        }
        model.con = Piecewise(*args, **keywords)

        try:
            keywords['pw_pts'] = [-1, 0, 0.5, 1]
            model.con3 = Piecewise(*args, **keywords)
        except Exception:
            pass
        else:
            self.fail(
                "Piecewise should fail when initialized "
                "with DLOG an pw_pts list with length not "
                "equal to (2^n)+1."
            )

    # test the that Piecewise component raises
    # an exception when the LOG/DLOG reps
    # are requested without a (2^n)+1 length
    # pw_pts list
    def test_log_bad_length(self):
        model = ConcreteModel()
        model.range = Var()

        model.x = Var(bounds=(-1, 1))
        args = (model.range, model.x)
        keywords = {
            'pw_pts': [-1, 0, 1],
            'pw_constr_type': 'EQ',
            'pw_repn': 'LOG',
            'f_rule': lambda model, x: x**2,
        }
        model.con = Piecewise(*args, **keywords)

        try:
            keywords['pw_pts'] = [-1, 0, 0.5, 1]
            model.con3 = Piecewise(*args, **keywords)
        except Exception:
            pass
        else:
            self.fail(
                "Piecewise should fail when initialized "
                "with LOG an pw_pts list with length not "
                "equal to (2^n)+1."
            )

    # test the that Piecewise component raises
    # an exception with an unsorted list of
    # domain points
    def test_unsorted_pw_pts(self):
        model = ConcreteModel()
        model.range = Var()

        model.x = Var(bounds=(-1, 1))
        args = (model.range, model.x)
        keywords = {
            'pw_pts': [-1, 0, 1],
            'pw_constr_type': 'EQ',
            'f_rule': lambda model, x: x**2,
        }
        model.con = Piecewise(*args, **keywords)

        try:
            keywords['pw_pts'] = [0, -1, 1]
            model.con3 = Piecewise(*args, **keywords)
        except Exception:
            pass
        else:
            self.fail(
                "Piecewise should fail when initialized with unsorted domain points."
            )

    # test the that Piecewise component raises
    # an exception when initialized without
    # a valid f_rule keyword
    def test_bad_f_rules(self):
        model = ConcreteModel()
        model.range = Var()

        model.x = Var(bounds=(-1, 1))
        args = (model.range, model.x)
        keywords = {
            'pw_pts': [-1, 0, 1],
            'pw_constr_type': 'EQ',
            'f_rule': lambda model, x: x**2,
        }
        model.con = Piecewise(*args, **keywords)

        try:
            del keywords['f_rule']
            model.con1 = Piecewise(*args, **keywords)
        except Exception:
            pass
        else:
            self.fail(
                "Piecewise should fail when initialized "
                "without a proper f_rule keyword."
            )

        try:
            keywords['f_rule'] = None
            model.con2 = Piecewise(*args, **keywords)
        except Exception:
            pass
        else:
            self.fail(
                "Piecewise should fail when initialized "
                "without a proper f_rule keyword."
            )

        try:
            keywords['f_rule'] = model.x
            model.con3 = Piecewise(*args, **keywords)
        except Exception:
            pass
        else:
            self.fail(
                "Piecewise should fail when initialized "
                "without a proper f_rule keyword."
            )

    # test the that Piecewise component raises
    # an exception if the domain variable arguments
    # are not actually Pyomo vars
    def test_bad_var_args(self):
        model = ConcreteModel()
        model.range = Var()

        model.x = Var(bounds=(-1, 1))
        args = (model.range, model.x)
        keywords = {
            'pw_pts': [-1, 0, 1],
            'pw_constr_type': 'EQ',
            'f_rule': lambda model, x: x**2,
        }
        model.con = Piecewise(*args, **keywords)
        try:
            args = (None, model.x)
            model.con1 = Piecewise(*args, **keywords)
        except Exception:
            pass
        else:
            self.fail(
                "Piecewise should fail when initialized "
                "without Pyomo vars as variable args."
            )

        try:
            args = (model.range, None)
            model.con2 = Piecewise(*args, **keywords)
        except Exception:
            pass
        else:
            self.fail(
                "Piecewise should fail when initialized "
                "without Pyomo vars as variable args."
            )

    # test the that Piecewise component raises
    # an exception if the piecewise bound type is
    # a bad value
    def test_bad_bound_type(self):
        model = ConcreteModel()
        model.range = Var()

        model.x = Var(bounds=(-1, 1))
        args = (model.range, model.x)
        keywords = {
            'pw_pts': [-1, 0, 1],
            'pw_constr_type': 'EQ',
            'f_rule': lambda model, x: x**2,
        }
        model.con = Piecewise(*args, **keywords)
        try:
            keywords['pw_constr_type'] = 1.0
            model.con1 = Piecewise(*args, **keywords)
        except Exception:
            pass
        else:
            self.fail("Piecewise should fail when initialized with invalid bound type.")
        try:
            del keywords['pw_constr_type']
            model.con1 = Piecewise(*args, **keywords)
        except Exception:
            pass
        else:
            self.fail("Piecewise should fail when initialized with invalid bound type.")

    # test the that Piecewise component raises
    # an exception if the piecewise representation
    # keyword has an invalid value
    def test_bad_repn(self):
        model = ConcreteModel()
        model.range = Var()

        model.x = Var(bounds=(-1, 1))
        args = (model.range, model.x)
        keywords = {
            'pw_pts': [-1, 0, 1],
            'pw_constr_type': 'EQ',
            'f_rule': lambda model, x: x**2,
        }
        model.con = Piecewise(*args, **keywords)
        try:
            keywords['pw_repn'] = 1.0
            model.con1 = Piecewise(*args, **keywords)
        except Exception:
            pass
        else:
            self.fail(
                "Piecewise should fail when initialized "
                "with invalid piecewise representation."
            )

    # test the that Piecewise component raises
    # an exception if the warning_tol
    # keyword has an invalid value
    def test_bad_warning_tol(self):
        model = ConcreteModel()
        model.range = Var()

        model.x = Var(bounds=(-1, 1))
        args = (model.range, model.x)
        keywords = {
            'pw_pts': [-1, 0, 1],
            'pw_constr_type': 'EQ',
            'f_rule': lambda model, x: x**2,
        }
        model.con = Piecewise(*args, **keywords)
        try:
            keywords['warning_tol'] = None
            model.con1 = Piecewise(*args, **keywords)
        except Exception:
            pass
        else:
            self.fail(
                "Piecewise should fail when initialized with invalid warning_tol."
            )

    # test the that Piecewise component raises
    # an exception if initialized with an invalid
    # number of args
    def test_bad_args_count(self):
        model = ConcreteModel()
        model.range = Var()

        model.x = Var(bounds=(-1, 1))
        args = (model.range, model.x)
        keywords = {
            'pw_pts': [-1, 0, 1],
            'pw_constr_type': 'EQ',
            'f_rule': lambda model, x: x**2,
        }
        model.con = Piecewise(*args, **keywords)
        try:
            args = (model.range,)
            model.con1 = Piecewise(*args, **keywords)
        except Exception:
            pass
        else:
            self.fail(
                "Piecewise should fail when initialized "
                "with less than two arguments."
            )

    # test the that Piecewise component raises
    # an exception if initialized with an unbounded
    # domain variable
    def test_unbounded_var(self):
        model = ConcreteModel()
        model.range = Var()

        model.x = Var(bounds=(-1, 1))
        args = (model.range, model.x)
        keywords = {
            'pw_pts': [-1, 0, 1],
            'pw_constr_type': 'EQ',
            'f_rule': lambda model, x: x**2,
        }
        model.con = Piecewise(*args, **keywords)
        try:
            model.x.setlb(None)
            model.x.setub(None)
            model.con1 = Piecewise(*args, **keywords)
        except Exception:
            pass
        else:
            self.fail(
                "Piecewise should fail when initialized "
                "with unbounded domain variable."
            )

        # Check that the unbounded_domain_var keyword works
        model.con1 = Piecewise(unbounded_domain_var=True, *args, **keywords)

        model.y = Var(bounds=(0, None))
        try:
            args = (model.range, model.y)
            model.con2 = Piecewise(*args, **keywords)
        except Exception:
            pass
        else:
            self.fail(
                "Piecewise should fail when initialized "
                "with unbounded domain variable."
            )

        model.z = Var(bounds=(None, 10))
        try:
            args = (model.range, model.z)
            model.con3 = Piecewise(*args, **keywords)
        except Exception:
            pass
        else:
            self.fail(
                "Piecewise should fail when initialized "
                "with unbounded domain variable."
            )

    def test_len(self):
        model = AbstractModel()
        model.range = Var()

        model.x = Var(bounds=(-1, 1))
        args = (model.range, model.x)
        keywords = {
            'pw_pts': [-1, 0, 1],
            'pw_constr_type': 'EQ',
            'f_rule': lambda model, x: x**2,
        }
        model.con = Piecewise(*args, **keywords)
        self.assertEqual(len(model.con), 0)
        instance = model.create_instance()
        self.assertEqual(len(instance.con), 1)

    def test_None_key(self):
        model = ConcreteModel()
        model.range = Var()

        model.x = Var(bounds=(-1, 1))
        args = (model.range, model.x)
        keywords = {
            'pw_pts': [-1, 0, 1],
            'pw_constr_type': 'EQ',
            'f_rule': lambda model, x: x**2,
        }
        model.con = Piecewise(*args, **keywords)
        self.assertEqual(id(model.con), id(model.con[None]))


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import sys
import typing

_overloads = {}


def _get_fullqual_name(func: typing.Callable) -> str:
    return f"{func.__module__}.{func.__qualname__}"


if sys.version_info[:2] <= (3, 10) and not typing.TYPE_CHECKING:

    def overload(func: typing.Callable):
        """Wrap typing.overload that remembers the overloaded signatures

        This provides a custom implementation of typing.overload that
        remembers the overloaded signatures so that they are available for
        runtime inspection.

        """
        _overloads.setdefault(_get_fullqual_name(func), []).append(func)
        return typing.overload(func)

    def get_overloads_for(func: typing.Callable):
        return _overloads.get(_get_fullqual_name(func), [])

else:
    from typing import get_overloads as get_overloads_for
    from typing import overload as overload
